from enum import Enum

from cyyrus.models.options import (
    ParsedFormat,
    UnparsedFormat,
)


class TaskType(str, Enum):
    # === Sentinel Tasks ===
    DEFAULT = "default"

    # === General Tasks ===
    PARSING = "parsing"  # Ingest and interpret data from different types.

    # === Multimodal Tasks ===
    GENERATION = (
        "generation"  # Create new content, such as generating text or audio from given inputs.
    )


TASK_PROPERTIES = {
    TaskType.DEFAULT: {
        "required": {},
        "optional": {},
    },
    TaskType.PARSING: {
        "required": {
            "parsed_format": (
                ParsedFormat,
                ParsedFormat.BASE64,
            ),
            "file_type": (
                UnparsedFormat,
                UnparsedFormat.PDF,
            ),
        },
        "optional": {
            "max_depth": (
                int,
                5,
            ),
            "directory": (
                str,
                "*",
            ),
        },
    },
    TaskType.GENERATION: {
        "required": {
            "prompt": (
                str,
                ...,
            ),
            "model": (
                str,
                ...,
            ),
            "api_key": (
                str,
                ...,
            ),
        },
        "optional": {
            "max_epochs": (
                int,
                100,
            ),
            "response_format": (
                str,
                None,
            ),
        },
    },
}



import pandas as pd
from jinja2 import Template

from cyyrus.models.spec import Spec


class MarkdownUtils:
    @staticmethod
    def generate_readme(
        spec: Spec,
        repository_id: str,
        dataframe: pd.DataFrame,
    ) -> str:
        template = Template(MarkdownUtils.get_template_string())
        return template.render(
            spec=spec,
            repository_id=repository_id,
            dataframe=dataframe,
        )

    @staticmethod
    def get_template_string():
        return """
# {{ spec.dataset.metadata.name or "Cyyrus Dataset" }}

{{ spec.dataset.metadata.description or "No description provided." }}

This dataset was generated using [Cyyrus](https://github.com/wizenheimer/cyyrus), an open-source library for creating and managing datasets.

## Using the Dataset

To use this dataset with the Hugging Face `datasets` library:

```python
from datasets import load_dataset

dataset = load_dataset("{{ repository_id or 'cyyrus/cyyrus-dataset' }}")
```

## License

This dataset is licensed under the {{ spec.dataset.metadata.license }} license.

## Tags

{% for tag in spec.dataset.metadata.tags %}
- {{ tag }}
{% endfor %}


## Acknowledgements

This dataset was created using [Cyyrus](https://github.com/wizenheimer/cyyrus), an open-source library for dataset generation and management. If you find yourself using this dataset in your work (and why wouldn't you?), consider giving Cyyrus a little love:

```bibtex
@software{cyyrus,
  title = {Cyyrus: An Open-Source Library for Dataset Generation and Management},
  author = {{cyyrus}},
  url = {https://github.com/wizenheimer/cyyrus},
  year = {2024}
}
```

For any questions or issues related to this dataset, please open an issue on the [Cyyrus GitHub repository](https://github.com/wizenheimer/cyyrus/issues).
"""



from abc import ABC, abstractmethod
from typing import Any, Dict, List

from cyyrus.models.task_type import TaskType
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class BaseTask(ABC):
    # The task ID is used to identify the task type
    TASK_ID = TaskType.DEFAULT

    # This flag indicates whether the task supports reference-free execution
    SUPPORTS_REFERENCE_FREE_EXECUTION = True

    def __init__(
        self,
        column_name: str,
        task_properties: Dict[str, Any],
    ) -> None:
        self.task_properties = task_properties
        self.column_name = column_name

    def reference_based_execution(
        self,
        task_input: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Perform a reference-based execution

        Strategy: Incase the task is reference-based, the task_input will be a dictionary containing the reference data. The final output will be a dictionary.
        """
        logger.debug(f"Executing reference based generation for task {self.TASK_ID} ...")
        interim_result = self.execute(
            task_input,
        )

        return {
            self.column_name: interim_result,
        }

    def reference_free_execution(
        self,
    ) -> List[Dict[str, Any]]:
        """
        Perform a reference-free execution.

        Strategy: Incase the task is reference-free, the task_input will be None and the task will attempt to generate the reference data. The final output will be a list of dictionaries.
        """
        logger.debug(f"Executing reference free generation for task {self.TASK_ID} ...")

        if not self.SUPPORTS_REFERENCE_FREE_EXECUTION:
            logger.error(f"Skipping, Task {self.TASK_ID} does not support reference-free execution")
            return []

        task_inputs = self._generate_references()
        task_results = []
        for task_input in task_inputs:
            task_results.append(self.reference_based_execution(task_input))
        return task_results

    # The execute method is the main method that needs to be implemented by the task
    @abstractmethod
    def execute(
        self,
        task_input: Dict[str, Any],
    ) -> Any:
        """
        Perform the task execution
        """
        logger.debug(f"Executing task {self.TASK_ID} ...")
        return None

    # Incase the task supports reference-free execution, the task should implement the _generate_references method as well
    def _generate_references(
        self,
    ) -> List[Dict[str, Any]]:
        """
        Attempt to generate the reference data for the task, using the task properties
        """
        return []



from enum import Enum


class UnparsedFormat(str, Enum):
    """
    Enum for the different formats in which data can be stored before parsing.
    """

    # Document Formats
    PDF = "pdf"

    # Image Formats
    PNG = "png"
    JPEG = "jpeg"
    JPG = "jpg"

    # Tabular Formats
    JSON = "json"
    CSV = "csv"


class ParsedFormat(str, Enum):
    """
    Enum for the different formats in which data can be stored after parsing.
    """

    BASE64 = "base64"
    MARKDOWN = "markdown"
    IMAGE = "image"


class LargeLanguageModels(str, Enum):
    """
    Enum for the different large language models supported by the API.
    """

    GPT_4 = "gpt-4"
    GPT_4O_MINI = "gpt-4o-mini"


class VisionLanguageModels(str, Enum):
    """
    Enum for the different vision language models supported by the API.
    """

    GPT_4 = "gpt-4"
    GPT_4O_MINI = "gpt-4o-mini"



from typing import Tuple

from datasets import Dataset

from cyyrus.constants.messages import Messages
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class DatasetUtils:
    @staticmethod
    def normalize_split_sizes(
        train_size: float,
        test_size: float,
    ) -> Tuple[float, float]:
        """
        Normalize the split sizes to ensure they add up to 1.
        """
        logger.debug(f"Validating splits: train: {train_size}, test: {test_size}")
        total = train_size + test_size
        if total <= 0:
            logger.debug(f"Old train split: {train_size}, Old test split: {test_size}")
            logger.debug("New train split: 1, New test split: 0")
            return 1.0, 0.0
        if train_size < 0:
            logger.debug(f"Old train split: {train_size}, Old test split: {test_size}")
            logger.debug("New train split: 0")
            train_size = 0
        if test_size < 0:
            logger.debug(f"Old train split: {train_size}, Old test split: {test_size}")
            logger.debug("New test split: 0")
            test_size = 0
        if abs(total - 1) > 1e-6:
            logger.warning(Messages.SPLITS_DONT_ADD_UP)
            logger.debug(f"Old train split: {train_size}, Old test split: {test_size}")
            logger.debug(
                f"New train split: {train_size / total}, New test split: {test_size / total}"
            )
            logger.debug(f"Normalizing splits: train: {train_size}, test: {test_size}")
            return train_size / total, test_size / total

        logger.debug(f"Final splits: train: {train_size}, test: {test_size}")
        return train_size, test_size

    @staticmethod
    def split_dataset(
        dataset: Dataset,
        train_size: float,
        test_size: float,
        seed: int,
    ) -> Tuple[Dataset, Dataset]:
        """
        Splits the dataset into train and test sets.
        """
        logger.debug(f"Splitting dataset with sizes: train: {train_size}, test: {test_size}")
        total_size = len(dataset)

        # Normalize split sizes
        train_size, test_size = DatasetUtils.normalize_split_sizes(
            train_size,
            test_size,
        )

        # Ensure at least one sample in each split
        min_samples = 1
        if total_size < 2 * min_samples:
            logger.warning(Messages.DATASET_TOO_SMALL_FOR_SPLIT)
            logger.warning(f"Total size: {total_size}, min samples: {min_samples}")
            return dataset, Dataset.from_dict({})

        train_samples = max(min_samples, int(total_size * train_size))
        test_samples = max(min_samples, total_size - train_samples)

        if train_samples + test_samples > total_size:
            logger.warning(Messages.RE_ADJUSTING_SPLIT)
            logger.warning(
                f"Train samples: {train_samples}, Test samples: {test_samples}, Total size: {total_size}"
            )
            if train_samples > test_samples:
                train_samples = total_size - min_samples
                test_samples = min_samples
            else:
                test_samples = total_size - min_samples
                train_samples = min_samples

        # Perform the split
        logger.debug(f"Performing dataset split {train_samples} train, {test_samples} test")
        split = dataset.train_test_split(
            train_size=train_samples, test_size=test_samples, seed=seed
        )
        return split["train"], split["test"]



import base64
from typing import Union


class Mermaid:
    """
    This class represents a Mermaid diagram and provides functionality
    to display it in a Jupyter notebook.
    """

    def __init__(self, diagram: str, position: Union[str, None] = None):
        """
        Initialize the Mermaid class with the Mermaid diagram script and optional position.

        Parameters:
            diagram (str): The Mermaid diagram script.
            position (Union[str, None]): The position of the diagram (left, right, center).
        """
        self.position = position
        self.diagram = self._process_diagram(diagram)
        self.svg_url = f"https://mermaid.ink/svg/{self.diagram}"

    def _process_diagram(self, diagram: str) -> str:
        """
        Process the Mermaid diagram script into a base64 encoded string.

        Parameters:
            diagram (str): The Mermaid diagram script.

        Returns:
            str: The base64 encoded string of the Mermaid diagram script.
        """
        graphbytes = diagram.encode("utf8")
        base64_bytes = base64.b64encode(graphbytes)
        return base64_bytes.decode("ascii")

    def _repr_html_(self) -> str:
        """
        Generate HTML to display the Mermaid diagram.

        Returns:
            str: HTML representation of the Mermaid diagram.
        """
        position_style = f"text-align:{self.position};" if self.position else ""
        return f'<div style="{position_style}"><img src="{self.svg_url}" /></div>'


# IPython magic function for easy use in Jupyter notebooks
# try:
#     from IPython.core.magic import register_cell_magic

#     @register_cell_magic
#     def mermaid(line, cell):
#         """
#         IPython cell magic for rendering Mermaid diagrams in Jupyter notebooks.

#         Parameters:
#             line (str): Options for rendering (e.g., --position).
#             cell (str): The Mermaid diagram script.
#         """
#         options = line.strip().split()
#         position = None
#         if "--position" in options:
#             position_index = options.index("--position") + 1
#             if position_index < len(options):
#                 position = options[position_index]
#         diagram = cell.strip()
#         mermaid = Mermaid(diagram, position)
#         display(HTML(mermaid._repr_html_()))

# except ImportError:
#     print("Error occurred while importing IPython modules.")



from enum import Enum
from typing import List, Optional, Tuple

from pydantic import BaseModel, Field, field_validator
from pydantic.functional_validators import model_validator

from cyyrus.constants.messages import Messages
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class SpecVersion(str, Enum):
    V0 = "v0"


class DatasetMetadata(BaseModel):
    name: str = Field(
        default="Dataset",
        description="Name of the dataset",
    )
    description: str = Field(
        default="Dataset generated using cyyrus",
        description="Description of the dataset",
    )
    tags: List[str] = Field(
        default=[
            "dataset",
            "cyyrus",
        ],
        description="Tags for the dataset",
    )
    license: str = Field(
        default="MIT",
        description="License for the dataset",
    )
    languages: List[str] = Field(
        default=[
            "en",
        ],
        description="Languages for the dataset",
    )

    @field_validator("tags")
    @classmethod
    def add_cyyrus_tag(
        cls,
        v: List[str],
    ) -> List[str]:
        if "cyyrus" not in v:
            v.append("cyyrus")
        return v


class DatasetShuffle(BaseModel):
    seed: int = Field(
        default=42,
        description="Seed for the random number generator",
    )


class DatasetSplits(BaseModel):
    train: Optional[float] = Field(
        default=None,
        ge=0,
        le=1,
        description="Percentage of the dataset to use for training",
    )
    test: Optional[float] = Field(
        default=None,
        ge=0,
        le=1,
        description="Percentage of the dataset to use for testing",
    )
    seed: Optional[int] = Field(
        default=42,
        description="Seed for the random number generator",
    )

    @model_validator(mode="after")
    def check_splits_sum(cls, values):
        logger.debug(f"Checking splits sum: {values}")
        train, test = values.train, values.test

        if train is None and test is None:
            logger.warn(Messages.SPLITS_DONT_ADD_UP)
            logger.debug("Both train and test splits are None, attempting to set default values")

            values.train, values.test = 0.8, 0.2

            logger.debug(f"Old train split: {train}, Old test split: {test}")
            logger.debug(f"New train split: {values.train}, New test split: {values.test}")

        elif train is None:
            logger.warn(Messages.SPLIT_VALUE_INVALID)
            logger.debug(f"Train split is None, attempting to compute from test split: {test}")

            values.train = max(0, 1 - test)

            logger.debug(f"Old train split: {train}, Old test split: {test}")
            logger.debug(f"New train split: {values.train}, New test split: {values.test}")
        elif test is None:
            logger.warn(Messages.SPLIT_VALUE_INVALID)
            logger.debug("Test split is None, attempting to compute from train split")

            values.test = max(0, 1 - train)

            logger.debug(f"Old train split: {train}, Old test split: {test}")
            logger.debug(f"New train split: {values.train}, New test split: {values.test}")

        logger.debug(f"Checking splits sum: {values}, train: {train}, test: {test}")
        normalized = cls.normalize_split_sizes(values.train, values.test)

        logger.debug(f"Normalized splits: {normalized}")
        values.train, values.test = normalized

        return values

    @staticmethod
    def normalize_split_sizes(
        train_size: float,
        test_size: float,
    ) -> Tuple[float, float]:
        total = train_size + test_size
        if total <= 0:
            logger.warning(f"{Messages.SPLITS_DONT_ADD_UP}")
            logger.warning(f"Old train split: {train_size}, Old test split: {test_size}")
            logger.warning(f"New train split: {0.8}, New test split: {0.2}")
            return 0.8, 0.2
        return round(train_size / total, 3), round(test_size / total, 3)

    @field_validator("train", "test")
    @classmethod
    def check_not_negative(cls, v: Optional[float]) -> Optional[float]:
        logger.debug(f"Checking if {v} is negative")
        if v is not None and v < 0:
            logger.warning(Messages.SPLIT_VALUE_INVALID)
            logger.debug(f"Old value: {v}, New value: {abs(v)}")
            return abs(v)
        return v


class DatasetAttributes(BaseModel):
    required_columns: List[str] = Field(
        default=[],
        description="Columns that are required in the dataset",
    )
    unique_columns: List[str] = Field(
        default=[],
        description="Columns that are unique in the dataset",
    )
    nulls: str = Field(
        default="include",
        pattern="^(include|exclude)$",
        description="How to handle null values",
    )
    flatten_columns: List[str] = Field(
        default=[],
        description="Columns that should be flattened",
    )
    exclude_columns: List[str] = Field(
        default=[],
        description="Columns that should be excluded",
    )


class Dataset(BaseModel):
    metadata: DatasetMetadata = Field(
        default_factory=DatasetMetadata,
        description="Metadata for the dataset",
    )
    shuffle: DatasetShuffle = Field(
        default_factory=DatasetShuffle,
        description="Shuffle configuration for the dataset",
    )
    splits: DatasetSplits = Field(
        default_factory=DatasetSplits,
        description="Splits for the dataset",
    )
    attributes: DatasetAttributes = Field(
        default_factory=DatasetAttributes,
        description="Attributes for the dataset",
    )



from enum import Enum
from typing import Any, Dict, List

import pandas as pd

from cyyrus.constants.messages import Messages
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class ExportFormat(str, Enum):
    HUGGINGFACE = "huggingface"
    JSON = "json"
    CSV = "csv"
    PICKLE = "pickle"
    PARQUET = "parquet"


class DataFrameUtils:
    @staticmethod
    def handle_nulls(
        df: pd.DataFrame,
        nulls: str,
    ) -> pd.DataFrame:
        """
        Drops rows with null values if nulls is set to "exclude".
        """
        if nulls == "exclude":
            return df.dropna()
        return df

    @staticmethod
    def ensure_required_columns(
        df: pd.DataFrame,
        required_columns: List[str],
    ) -> pd.DataFrame:
        """
        Ensure that the required columns are present in the DataFrame.
        """
        missing_columns = set(required_columns) - set(df.columns)
        if missing_columns:
            logger.warning(Messages.REQUIRED_COLUMN_MISSING)
            logger.warning(f"Missing columns: {missing_columns}")
            for col in missing_columns:
                df[col] = pd.Series(dtype=object)
        return df

    @staticmethod
    def ensure_unique_columns(
        df: pd.DataFrame,
        unique_columns: List[str],
    ) -> pd.DataFrame:
        """
        Ensure that the unique columns have unique values in the DataFrame.
        Drops duplicates if unique_columns is not empty.
        """
        if unique_columns:
            initial_count = len(df)
            df = df.drop_duplicates(subset=unique_columns, keep="first")
            removed_count = initial_count - len(df)
            if removed_count > 0:
                logger.warning(Messages.NON_UNIQUE_COLUMN_VALUES)
                logger.warning(f"Unique columns: {unique_columns}, Removed count: {removed_count}")
        return df

    @staticmethod
    def handle_flattening(
        df: pd.DataFrame,
        columns_to_flatten: List[str],
    ) -> pd.DataFrame:
        """
        Robustly flatten specified dictionary columns in a DataFrame.

        Args:
        df (pd.DataFrame): The input DataFrame.
        columns_to_flatten (list): List of column names containing dictionaries to flatten.

        Returns:
        pd.DataFrame: A new DataFrame with specified dictionary columns flattened.
        """
        df = df.copy()
        flattened_columns = []

        for col in columns_to_flatten:
            if col not in df.columns:
                logger.debug(f"Warning: Column '{col}' not found in the DataFrame. Skipping.")
                continue

            if not df[col].dtype == "object":
                logger.debug(f"Warning: Column '{col}' is not of object type. Skipping.")
                continue

            # Check if the column contains dictionaries
            if not df[col].apply(lambda x: isinstance(x, dict)).all():
                logger.debug(
                    f"Warning: Column '{col}' does not contain only dictionaries. Skipping."
                )
                continue

            try:
                # Flatten the dictionary column
                flattened = pd.json_normalize(df[col].apply(lambda x: x or {}))  # type: ignore

                # Rename the new columns to avoid conflicts
                flattened.columns = [f"{col}_{key}" for key in flattened.columns]

                # Add the flattened columns to our list
                flattened_columns.append(flattened)

                # Drop the original column
                df = df.drop(columns=[col])
            except Exception as e:
                logger.debug(f"Error flattening column '{col}': {str(e)}. Skipping.")

        # Concatenate all flattened columns with the original DataFrame
        if flattened_columns:
            df = pd.concat([df] + flattened_columns, axis=1)

        return df

    # Helper function to safely get nested dictionary values
    @staticmethod
    def safe_get(
        dct: Dict[str, Any],
        *keys: str,
    ) -> Any:
        for key in keys:
            try:
                dct = dct[key]
            except (KeyError, TypeError):
                return None
        return dct

    @staticmethod
    def remove_columns(
        df: pd.DataFrame,
        columns_to_remove: List[str],
    ) -> pd.DataFrame:
        """
        Remove specified columns from a DataFrame if they exist.

        Args:
        df (pd.DataFrame): The input DataFrame.
        columns_to_remove (List[str]): List of column names to remove.

        Returns:
        pd.DataFrame: A new DataFrame with specified columns removed.
        """
        df = df.copy()
        existing_columns = df.columns.tolist()

        for col in columns_to_remove:
            if col in existing_columns:
                df = df.drop(columns=[col])
                logger.debug(f"Removed column: '{col}'")
            else:
                logger.debug(f"Column '{col}' not found in the DataFrame. Skipping.")

        return df



import importlib
import inspect
import tempfile
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

import pandas as pd
from datasets import Dataset, DatasetDict
from huggingface_hub.hf_api import HfApi

from cyyrus.composer.dataframe import (
    DataFrameUtils,
    ExportFormat,
)
from cyyrus.composer.dataset import DatasetUtils
from cyyrus.composer.markdown import MarkdownUtils
from cyyrus.composer.progress import conditional_tqdm
from cyyrus.constants.messages import Messages
from cyyrus.models.spec import Spec
from cyyrus.models.task import TaskType
from cyyrus.tasks.base import BaseTask
from cyyrus.tasks.default import DefaultTask
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class Composer:
    def __init__(self, spec: Spec) -> None:
        logger.debug("Initializing Composer with Spec")
        self.spec: Spec = spec
        self.dataframe: pd.DataFrame = pd.DataFrame()

        logger.debug("Infering task artifacts")
        self.task_artifacts = self._infer_tasks()

        logger.debug(f"Discovered {len(self.task_artifacts)} total possible tasks")
        logger.debug("Composer initialized")

    def _infer_tasks(self) -> Dict[TaskType, Type[BaseTask]]:
        try:
            logger.debug("Importing tasks module")
            module = importlib.import_module("cyyrus.tasks")
        except ImportError as _:
            raise
        else:
            task_dict: Dict[TaskType, Type[BaseTask]] = {}

            for _, cls in inspect.getmembers(module, inspect.isclass):
                if hasattr(cls, "TASK_ID") and (issubclass(cls, BaseTask)) and cls != BaseTask:
                    logger.debug(f"Registering task: {cls.TASK_ID} to task artifacts")
                    task_dict[cls.TASK_ID] = cls
                else:
                    logger.debug(f"Skipping class: {cls} from task artifacts")

            return task_dict

    def compose(
        self,
        dry_run: bool = False,
    ):
        logger.debug("Iterating over levels in the spec")
        # Iterate over each level in the spec
        for level_index, level in enumerate(self.spec.levels()):
            logger.debug(f"Processing level: {level_index}")
            # Merge tasks in the level
            for input_columns, output_column, task_type, task_properties in level:
                logger.debug(f"Processing task: {task_type}")

                if dry_run:
                    logger.info("Dry run enabled, skipping execution")
                    logger.info(f"Task: {task_type}")
                    logger.info(f"Inputs: {input_columns}")
                    logger.info(f"Output: {output_column}")
                    continue

                self.execute(
                    input_columns=input_columns,
                    output_column=output_column,
                    task_type=task_type,
                    task_properties=task_properties,
                    level_index=level_index,
                )

    def execute(
        self,
        input_columns: List[str],
        output_column: str,
        task_type: TaskType,
        task_properties: Dict[str, Any],
        level_index: int = 0,
        dry_run: bool = False,
    ):
        logger.info(f"Preparing column: {output_column}")
        logger.info(f"Executing task: {task_type}")
        logger.debug(f"Inputs: {input_columns}")
        logger.debug(f"Output: {output_column}")

        task_entity = self.task_artifacts.get(task_type, DefaultTask)
        task_instance = task_entity(
            column_name=output_column,
            task_properties=task_properties,
        )

        if dry_run:
            logger.info("Dry run enabled, skipping execution")
            logger.info(f"Task: {task_type}")
            logger.info(f"Input Columns: {input_columns[:2]}...")
            logger.info(f"Output Columns: {output_column}")
            return

        task_inputs = self._import_columns(columns=input_columns)

        task_results: List[Dict[str, Any]] = []
        # Incase there are no task_inputs we attempt reference free execution
        if not task_inputs and level_index == 0:
            logger.debug("No task inputs, attempting reference free execution")
            task_results: List[Dict[str, Any]] = task_instance.reference_free_execution()
        # Incase there are task_inputs we attempt reference based execution
        else:
            logger.debug("Task inputs found, attempting reference based execution")
            logger.debug(f"Total Task inputs: {len(task_inputs)}")
            for task_input in conditional_tqdm(
                task_inputs,
                use_tqdm=not dry_run,
            ):
                task_output = task_instance.reference_based_execution(task_input)
                task_results.append({**task_input, **task_output})

        self._refresh_dataframe(
            column_data=task_results,
        )

    def export(
        self,
        export_format: Union[ExportFormat, str],
        filepath: Union[str, Path],
    ) -> None:
        logger.debug(f"Exporting data in {export_format} format to {filepath}")

        # Validate and convert format to ExportFormat enum
        if isinstance(export_format, str):
            try:
                export_format = ExportFormat(export_format.lower())
            except ValueError:
                raise ValueError(f"Unsupported export format: {export_format}")

        # Validate and convert filepath to Path
        if isinstance(filepath, str):
            filepath = Path(filepath)

        # Ensure the directory exists
        filepath.parent.mkdir(parents=True, exist_ok=True)

        # Check if file already exists
        if filepath.exists():
            logger.warning(f"File {filepath} already exists. It will be overwritten.")

        try:
            # Prepare the data
            prepared_data = self._prepare(export_format)

            # Export the data
            if isinstance(prepared_data, DatasetDict) or isinstance(prepared_data, Dataset):
                logger.debug("Exporting Hugging Face DatasetDict")
                prepared_data.save_to_disk(filepath)
                readme_content = MarkdownUtils.generate_readme(
                    self.spec,
                    self.spec.dataset.metadata.name,
                    self.dataframe,
                )
                with open(f"{filepath}/README.md", "w") as readme_file:
                    readme_file.write(readme_content)

            elif isinstance(prepared_data, pd.DataFrame):
                if export_format == ExportFormat.JSON:
                    prepared_data.to_json(
                        filepath,
                        orient="records",
                        lines=True,
                    )
                elif export_format == ExportFormat.CSV:
                    prepared_data.to_csv(
                        filepath,
                        index=False,
                    )
                elif export_format == ExportFormat.PICKLE:
                    prepared_data.to_pickle(
                        filepath,
                    )
                elif export_format == ExportFormat.PARQUET:
                    prepared_data.to_parquet(
                        filepath,
                    )
            else:
                raise TypeError(f"Unsupported data type for exporting: {type(prepared_data)}")

            logger.info(f"Data successfully exported to {filepath}")
        except Exception as e:
            logger.error(f"Failed to export data: {str(e)}")
            raise

    # Prepare the dataframe for export
    def _prepare(
        self,
        export_format: ExportFormat = ExportFormat.HUGGINGFACE,
    ) -> Union[DatasetDict, pd.DataFrame]:
        logger.debug(f"Preparing dataframe for {export_format} export")

        logger.debug("Handling flattening")
        df = DataFrameUtils.handle_flattening(
            self.dataframe,
            self.spec.dataset.attributes.flatten_columns,
        )

        # Handle nulls
        logger.debug("Attempting to handle nulls")
        df = DataFrameUtils.handle_nulls(
            self.dataframe,
            self.spec.dataset.attributes.nulls,
        )

        # Ensure unique columns
        logger.debug("Ensuring required columns")
        df = DataFrameUtils.ensure_required_columns(
            df,
            self.spec.dataset.attributes.required_columns,
        )

        # Ensure required columns
        logger.debug("Ensuring required columns")
        df = DataFrameUtils.ensure_unique_columns(
            df,
            self.spec.dataset.attributes.unique_columns,
        )

        # Remove unnecessary columns
        logger.debug("Removing unnecessary columns")
        df = DataFrameUtils.remove_columns(
            df=df,
            columns_to_remove=self.spec.dataset.attributes.exclude_columns,
        )

        if export_format == ExportFormat.HUGGINGFACE:
            # Convert to Hugging Face Dataset
            logger.debug("Converting dataframe to Hugging Face Dataset")
            dataset = Dataset.from_pandas(df)

            # Shuffle the dataset
            logger.debug("Shuffling the dataset")
            dataset = dataset.shuffle(seed=self.spec.dataset.shuffle.seed)

            # Split the dataset
            logger.debug("Splitting the dataset")
            train_set, test_set = DatasetUtils.split_dataset(
                dataset,
                self.spec.dataset.splits.train or 0.8,  # default to 0.8 if not specified
                self.spec.dataset.splits.test or 0.2,  # default to 0.2 if not specified
                self.spec.dataset.splits.seed or 42,  # default to 42 if not specified
            )

            # Create DatasetDict
            logger.debug("Creating DatasetDict with train and test splits")
            hf_dataset = DatasetDict(
                {
                    "train": train_set,
                    "test": test_set,
                },
            )

            # Set dataset metadata
            logger.debug("Setting dataset metadata")
            for split in hf_dataset.values():
                split.info.description = self.spec.dataset.metadata.description
                split.info.license = self.spec.dataset.metadata.license

            logger.debug("Export complete")
            return hf_dataset

        else:
            logger.debug(f"Preparing dataframe for {format} format")
            return df

    def _import_columns(
        self,
        columns: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        This method exports column names from the dataframe. It:

        1. Exports all column names if column_names is None
        2. Exports only the specified column names if column_names is provided
        3. Returns the exported column names
        """
        logger.debug(f"Importing columns {columns[:2] if columns else None} from the dataframe")

        # Check if the dataframe is empty or no columns are specified
        if self.dataframe.empty or not columns:
            logger.debug("Dataframe is empty or no columns specified")
            return []

        # Check if all column names are valid
        logger.debug("Checking if all column names are valid")
        invalid_column_names = set(columns) - set(self.dataframe.columns)
        if invalid_column_names:
            # TODO: attempt to find the closest matching column names
            logger.error(Messages.INVALID_KEY_COLUMN)
            logger.error(f"Invalid column names: {invalid_column_names}")
            logger.error(f"Valid column names: {self.dataframe.columns}")
            raise ValueError(Messages.INVALID_KEY_COLUMN)

        # Export only the specified columns
        logger.debug("Exporting only the specified columns")
        result = [
            {k: v for k, v in row.items() if k in columns}
            for row in self.dataframe[columns].to_dict("records")  # type: ignore
            if all(pd.notna(v) for v in row.values())
        ]

        # Check if all rows were excluded due to NaN values
        logger.debug("Checking if all rows were excluded due to NaN values")
        if not result:
            logger.warning(Messages.ALL_ROWS_EXCLUDED_DUE_TO_NAN)

        logger.debug(f"Imported {len(result)} rows")
        return result

    def _refresh_dataframe(self, column_data: List[Dict[str, Any]]):
        """
        Refreshes the dataframe with new column data, handling potential duplicate indices and preserving existing data.

        :param column_data: List of dictionaries containing new data

        Strategy:
        If a column exists in both dataframes, we update it using combine_first.
        If a column only exists in the old dataframe, we keep it as is.
        If a column only exists in the new dataframe, we add it.
        """
        logger.debug("Refreshing dataframe with new column data")

        # Handle empty column data
        logger.debug("Handling empty column data")
        if not column_data:
            return

        # Convert column data to a DataFrame
        logger.debug("Converting column data to DataFrame")
        new_df = pd.DataFrame(column_data)

        # Explode list columns in new_df
        logger.debug("Exploding list columns in new dataframe")
        for col in new_df.columns:
            if new_df[col].apply(lambda x: isinstance(x, list)).any():
                new_df = new_df.explode(col)

        # Reset index of new_df to ensure it's unique
        new_df = new_df.reset_index(drop=True)

        # If the existing dataframe is empty, just use the new data
        logger.debug("Checking if the existing dataframe is empty")

        if self.dataframe.empty:
            logger.debug("Existing dataframe is empty, swapping previous dataframe")
            self.dataframe = new_df
            return

        # Reset index of existing dataframe to ensure it's unique
        logger.debug("Resetting index of existing dataframe")
        self.dataframe = self.dataframe.reset_index(drop=True)

        # Identify new columns
        logger.debug("Identifying new columns")
        existing_columns = set(self.dataframe.columns)

        # Create a temporary dataframe with all columns
        logger.debug("Creating a temporary dataframe with all columns")
        temp_df = pd.DataFrame(index=range(max(len(self.dataframe), len(new_df))))

        # Update existing columns and add new ones
        logger.debug("Updating existing columns and adding new ones")
        for col in self.dataframe.columns.union(new_df.columns):
            if col in existing_columns and col in new_df.columns:
                # Update existing column, preserving old values where new ones are NaN
                temp_df[col] = new_df[col].combine_first(self.dataframe[col])
            elif col in existing_columns:
                # Keep existing column as is
                temp_df[col] = self.dataframe[col]
            else:
                # Add new column
                temp_df[col] = new_df[col]

        # Update self.dataframe with the merged data
        logger.debug("Updating self.dataframe with the merged data")
        self.dataframe = temp_df.dropna(how="all").reset_index(drop=True)

        logger.debug("Dataframe refreshed")

    def publish(
        self,
        repository_id: str,
        huggingface_token: str,
        private: bool = False,
    ):
        logger.info(f"Publishing dataset to Hugging Face: {repository_id}")

        try:
            # Prepare the dataset
            hf_dataset: DatasetDict = self._prepare(ExportFormat.HUGGINGFACE)  # type: ignore

            # Initialize Hugging Face API
            api = HfApi(token=huggingface_token)

            # Create or update the dataset on Hugging Face
            api.create_repo(
                repo_id=repository_id,
                repo_type="dataset",
                exist_ok=True,
                private=private,
            )

            # # Save the dataset to a temporary directory
            with tempfile.TemporaryDirectory() as tmp_dir:
                # Save the dataset to the temporary directory
                hf_dataset.save_to_disk(tmp_dir)

                # Generate the README file
                readme_content = MarkdownUtils.generate_readme(
                    self.spec,
                    repository_id,
                    self.dataframe,
                )
                with open(f"{tmp_dir}/README.md", "w") as readme_file:
                    readme_file.write(readme_content)

                # Upload the dataset files
                api.upload_folder(
                    repo_id=repository_id,
                    folder_path=tmp_dir,
                    repo_type="dataset",
                    multi_commits=True,
                )

            logger.info(f"Dataset successfully published to {repository_id}")
        except Exception as e:
            logger.error(f"Failed to publish dataset: {str(e)}")
            raise



from typing import get_args, get_origin

import pytest
from cyyrus.errors.types import MaximumDepthExceededError  # type: ignore
from cyyrus.models.types import (  # type: ignore
    ArrayItems,
    CustomType,
    DataType,
    ObjectProperty,
    get_dynamic_model,
)
from pydantic import BaseModel, ValidationError


@pytest.mark.xfail(reason="Alias types are not supported in the schema for v0")
def test_alias_types():
    # Test with custom string type
    items2 = ArrayItems(type="custom_type")  # type: ignore
    assert items2.type == "custom_type"
    assert items2.properties is None

    # Test with custom string type
    prop2 = ObjectProperty(type="custom_type")  # type: ignore
    assert prop2.type == "custom_type"

    # Test with invalid type
    CustomType(type="invalid_type")


def test_array_items():
    # Test valid ArrayItems
    valid_items = ArrayItems(type=DataType.STRING)
    assert valid_items.type == DataType.STRING
    assert valid_items.properties is None

    # Test with DataType
    items1 = ArrayItems(type=DataType.INTEGER)
    assert items1.type == DataType.INTEGER
    assert items1.properties is None

    # Test with properties
    items3 = ArrayItems(
        type=DataType.OBJECT,
        properties={"prop1": ObjectProperty(type=DataType.STRING), "prop2": "integer"},
    )
    assert items3.type == DataType.OBJECT
    assert items3.properties["prop1"].type == DataType.STRING  # type: ignore
    assert items3.properties["prop2"].type == "integer"  # type: ignore

    # Test ArrayItems with properties
    items_with_props = ArrayItems(
        type=DataType.OBJECT, properties={"prop1": ObjectProperty(type=DataType.INTEGER)}
    )
    assert items_with_props.type == DataType.OBJECT
    assert "prop1" in items_with_props.properties  # type: ignore
    assert items_with_props.properties["prop1"].type == DataType.INTEGER  # type: ignore

    # Test ArrayItems with string property
    items_with_string_prop = ArrayItems(type=DataType.OBJECT, properties={"prop1": "string"})
    assert items_with_string_prop.properties["prop1"].type == "string"  # type: ignore

    # Test with invalid type
    with pytest.raises(ValidationError):
        ArrayItems(type=123)  # type: ignore

    # Test with invalid properties
    with pytest.raises(ValidationError):
        ArrayItems(type=DataType.OBJECT, properties={"prop1": 123})  # type: ignore


def test_type():
    # Test valid Type
    valid_type = CustomType(type=DataType.STRING)
    assert valid_type.type == DataType.STRING
    assert valid_type.properties is None
    assert valid_type.items is None

    # Test Type with properties
    type_with_props = CustomType(
        type=DataType.OBJECT, properties={"prop1": ObjectProperty(type=DataType.INTEGER)}
    )
    assert type_with_props.type == DataType.OBJECT
    assert "prop1" in type_with_props.properties  # type: ignore
    assert type_with_props.properties["prop1"].type == DataType.INTEGER  # type: ignore

    # Test Type with items
    type_with_items = CustomType(type=DataType.ARRAY, items=ArrayItems(type=DataType.STRING))
    assert type_with_items.type == DataType.ARRAY
    assert type_with_items.items.type == DataType.STRING  # type: ignore

    # Test invalid Type (missing required field)
    with pytest.raises(ValidationError):
        CustomType()

    # Test Type with invalid dynamic model
    with pytest.raises(ValidationError):
        CustomType(
            type="invalid_type"
        )  # Deviates from the schema, which doesn't allow custom/aliased types


def test_create_dynamic_type():
    # Test primitive types
    str_model = get_dynamic_model({"type": "string"})
    assert issubclass(str_model, BaseModel)
    assert str_model.model_fields["value"].annotation is str

    int_model = get_dynamic_model({"type": "integer"})
    assert issubclass(int_model, BaseModel)
    assert int_model.model_fields["value"].annotation is int

    float_model = get_dynamic_model({"type": "float"})
    assert issubclass(float_model, BaseModel)
    assert float_model.model_fields["value"].annotation is float

    bool_model = get_dynamic_model({"type": "boolean"})
    assert issubclass(bool_model, BaseModel)
    assert bool_model.model_fields["value"].annotation is bool

    # Test object type
    obj_model = get_dynamic_model(
        {
            "type": "object",
            "properties": {"prop1": {"type": "string"}, "prop2": {"type": "integer"}},
        }
    )

    assert issubclass(obj_model, BaseModel)
    assert issubclass(obj_model.model_fields["prop1"].annotation, BaseModel)  # type: ignore
    assert obj_model.model_fields["prop1"].annotation.model_fields["value"].annotation is str
    assert issubclass(obj_model.model_fields["prop2"].annotation, BaseModel)  # type: ignore
    assert obj_model.model_fields["prop2"].annotation.model_fields["value"].annotation is int

    # Test array type
    array_model = get_dynamic_model({"type": "array", "items": {"type": "string"}})
    assert issubclass(array_model, BaseModel)
    items_type = array_model.model_fields["items"].annotation
    assert get_origin(items_type) is list
    item_type = get_args(items_type)[0]
    assert issubclass(item_type, BaseModel)
    assert item_type.model_fields["value"].annotation is str

    # Test nested object type
    nested_obj_model = get_dynamic_model(
        {
            "type": "object",
            "properties": {
                "nested": {
                    "type": "object",
                    "properties": {"prop1": {"type": "string"}, "prop2": {"type": "integer"}},
                }
            },
        }
    )

    assert issubclass(nested_obj_model, BaseModel)
    assert issubclass(nested_obj_model.model_fields["nested"].annotation, BaseModel)  # type: ignore
    nested_model = nested_obj_model.model_fields["nested"].annotation
    assert issubclass(nested_model.model_fields["prop1"].annotation, BaseModel)  # type: ignore
    assert nested_model.model_fields["prop1"].annotation.model_fields["value"].annotation is str
    assert issubclass(nested_model.model_fields["prop2"].annotation, BaseModel)  # type: ignore
    assert nested_model.model_fields["prop2"].annotation.model_fields["value"].annotation is int

    # Test maximum depth exceeded
    with pytest.raises(MaximumDepthExceededError) as excinfo:
        get_dynamic_model(
            {
                "type": "object",
                "properties": {
                    "nested1": {
                        "type": "object",
                        "properties": {
                            "nested2": {
                                "type": "object",
                                "properties": {
                                    "nested3": {
                                        "type": "object",
                                        "properties": {
                                            "nested4": {
                                                "type": "object",
                                                "properties": {"nested5": {"type": "string"}},
                                            }
                                        },
                                    }
                                },
                            }
                        },
                    }
                },
            },
        )
    assert excinfo.value.extra_info["max_depth"] == "5"

    # Test with custom max_depth
    deep_nested_model = get_dynamic_model(
        {
            "type": "object",
            "properties": {
                "nested1": {
                    "type": "object",
                    "properties": {
                        "nested2": {
                            "type": "object",
                            "properties": {"prop": {"type": "string"}},
                        }
                    },
                }
            },
        },
        max_depth=10,
    )
    assert issubclass(deep_nested_model, BaseModel)
    assert issubclass(deep_nested_model.model_fields["nested1"].annotation, BaseModel)  # type: ignore
    nested1_model = deep_nested_model.model_fields["nested1"].annotation
    assert issubclass(nested1_model.model_fields["nested2"].annotation, BaseModel)  # type: ignore
    nested2_model = nested1_model.model_fields["nested2"].annotation
    assert issubclass(nested2_model.model_fields["prop"].annotation, BaseModel)  # type: ignore
    assert nested2_model.model_fields["prop"].annotation.model_fields["value"].annotation is str



import re
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import (
    BaseModel,
    create_model,
)
from pydantic.fields import Field
from pydantic.functional_validators import model_validator

from cyyrus.constants.messages import Messages
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class DataType(str, Enum):
    """
    Enum for the different data types supported by the API.
    """

    DEFAULT = "string"
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    BOOLEAN = "boolean"
    OBJECT = "object"
    ARRAY = "array"


class ObjectProperty(BaseModel):
    """
    Pydantic model for defining an object property.
    """

    type: DataType


class ArrayItems(BaseModel):
    """
    Pydantic model for defining the items in an array.
    """

    type: DataType
    properties: Optional[Dict[str, Union[ObjectProperty, str]]] = None

    @model_validator(mode="after")
    def validate_properties(cls, values):
        logger.debug(f"Validating properties for {cls}")
        if values.properties:
            for key, value in values.properties.items():
                if isinstance(value, str):
                    values.properties[key] = ObjectProperty(
                        type=DataType(value),
                    )
        logger.debug(f"Properties after validation: {values.properties}")
        return values


class CustomType(BaseModel):
    type: DataType
    properties: Optional[
        Dict[
            str,
            Union[
                str,
                ObjectProperty,
                Dict[str, Any],
            ],
        ]
    ] = None
    items: Optional[ArrayItems] = None

    @model_validator(mode="after")
    def validate_properties(cls, values):
        logger.debug(f"Validating properties for {cls}")
        if values.properties:
            for key, value in values.properties.items():
                if isinstance(value, str):
                    values.properties[key] = ObjectProperty(type=value)  # type: ignore
        logger.debug(f"Properties after validation: {values.properties}")
        return values

    @model_validator(mode="after")
    def validate_items(cls, values):
        logger.debug(f"Validating items for {cls}")
        if values.items and isinstance(values.items, dict):
            values.items = ArrayItems(**values.items)
        logger.debug(f"Items after validation: {values.items}")
        return values


# ================================================================================================
#                                   Type Mapping Functions
# These functions take task and type definitions and return the appropriate type as pydantic model
# ================================================================================================


class DefaultModel(BaseModel):
    """
    Pydantic model for the default data type.
    """

    value: str = Field(..., description="value")


class MarkdownModel(BaseModel):
    """
    Pydantic model representing the structure and content of a Markdown file.
    """

    markdown: str = Field(..., description="Markdown content")


class TypeMappingUtils:
    """
    Utility class for mapping task and type definitions to a concrete pydantic models.
    """

    @staticmethod
    def sanitize_name(name: str) -> str:
        """
        Sanitize the given name to ensure it only contains valid characters.
        """
        # Replace any non-alphanumeric characters with underscores
        sanitized = re.sub(r"[^a-zA-Z0-9]", "_", name)
        # Ensure the name starts with a letter
        if not sanitized[0].isalpha():
            sanitized = "Model_" + sanitized
        return sanitized

    @staticmethod
    def get_python_type(
        type_string: str,
    ) -> type:
        """
        Get the python type corresponding to the given type string.
        """

        TYPE_MAPPING = {
            "string": str,
            "integer": int,
            "float": float,
            "boolean": bool,
        }

        return TYPE_MAPPING.get(
            type_string.lower(),
            Any,
        )

    @staticmethod
    def get_concrete_model(
        type_def: Optional[Dict[str, Any]] = None,
        depth: int = 0,
        max_depth: int = 5,
    ) -> type[BaseModel]:
        """
        Get the concrete pydantic model for the given type definition.
        """

        if type_def is None:
            logger.debug("Type definition is None, using default model")
            return DefaultModel

        # Check if the depth exceeds the maximum depth
        if depth >= max_depth:
            logger.error(f"{Messages.MAXIMUM_DEPTH_EXCEEDED}")
            logger.debug(f"Max depth: {max_depth}, Current depth: {depth}")
            raise ValueError(Messages.MAXIMUM_DEPTH_EXCEEDED)

        # Get the type value from the type definition
        type_value = type_def["type"]

        logger.debug(f"Creating model for type value: {type_value}")

        # Recursively create models for nested types
        if type_value == "object":
            logger.debug("Creating model for object type")
            model_name = TypeMappingUtils.sanitize_name(f"ObjectModel_{depth}")

            fields = {}

            # Iterate over the properties and create models for each property
            for prop_name, prop_type in type_def.get(
                "properties",
                {},
            ).items():
                # If the property type is a dictionary, create a model for it
                if isinstance(prop_type, dict):
                    fields[prop_name] = (
                        TypeMappingUtils.get_concrete_model(
                            prop_type,
                            depth + 1,
                            max_depth,
                        ),
                        ...,
                    )
                else:
                    fields[prop_name] = (
                        TypeMappingUtils.get_python_type(
                            prop_type,
                        ),
                        ...,
                    )
            return create_model(
                model_name,
                **fields,
            )
        # If the type is an array, create a model for the items
        elif type_value == "array":
            logger.debug("Creating model for array type")
            model_name = TypeMappingUtils.sanitize_name(f"ArrayModel_{depth}")

            item_type = TypeMappingUtils.get_concrete_model(
                type_def["items"],
                depth + 1,
                max_depth,
            )
            return create_model(
                model_name,
                items=(
                    List[item_type],
                    ...,
                ),
            )
        # If the type is a primitive type, return the corresponding python type
        else:
            logger.debug("Creating model for primitive type")
            model_name = TypeMappingUtils.sanitize_name(f"{type_value.capitalize()}Model_{depth}")

            python_type = TypeMappingUtils.get_python_type(
                type_value,
            )
            return create_model(
                model_name,
                value=(python_type, ...),
            )



import base64
import copy
import inspect
import os
import warnings
from functools import lru_cache
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from openai import OpenAI
from pdf2image import convert_from_path
from PIL import Image
from pydub import AudioSegment

from cyyrus.models.options import ParsedFormat
from cyyrus.models.task_type import TaskType
from cyyrus.models.types import MarkdownModel
from cyyrus.tasks.base import BaseTask
from cyyrus.tasks.utils import Base64ImageFinder, GeneralUtils, NestedDictAccessor
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


# For reference based generation, there's always a one-to-one mapping between the input and output
# For instance, for a given PDF file, we can generate multiple images, so the resulting output would be a list of images encoded in base64
# For reference free generation, there's no such mapping constraint
# Hence, the output would be a single image encoded in base64
class ParsingTask(BaseTask):
    # The task ID is used to identify the task type
    TASK_ID = TaskType.PARSING

    # This flag indicates whether the task supports reference-free execution
    SUPPORTS_REFERENCE_FREE_EXECUTION = True

    # Default values for task properties
    DEFAULT_DIRECTORY = str(Path.cwd())
    DEFAULT_FILE_TYPE = "pdf"
    DEFAULT_MAX_DEPTH = 5
    DEFAULT_PARSED_FORMAT = ParsedFormat.BASE64

    EXPECTED_KEY = "path"

    DEFAULT_PROMPT = """Convert the following image to markdown.
    Return only the markdown with no explanation text.
    Do not exclude any content from the image."""

    # At present, we support parsing of PDF and image files
    DOCUMENT_TYPE = [
        "pdf",
    ]
    IMAGE_TYPE = [
        "png",
        "jpg",
        "jpeg",
    ]

    def __init__(
        self,
        column_name: str,
        task_properties: Dict[str, Any],
    ) -> None:
        """
        Initialize the parsing task.
        """
        super().__init__(
            column_name,
            task_properties,
        )
        # Set the default prompt in case it is not provided
        self.task_properties["prompt"] = (
            ParsingTask.DEFAULT_PROMPT
            if "prompt" not in self.task_properties
            else self.task_properties["prompt"]
        )

        parsed_format = NestedDictAccessor.get_nested_value(
            data=self.task_properties,
            key="parsed_format",
            default=ParsingTask.DEFAULT_PARSED_FORMAT,
        )

        if parsed_format == ParsedFormat.MARKDOWN:
            api_key = NestedDictAccessor.get_nested_value(
                data=self.task_properties,
                key="api_key",
                default=None,
            )
            self.client = OpenAI(api_key=api_key)

    def execute(
        self,
        task_input: Dict[str, Any],
    ) -> Any:
        """
        Perform the parsing task.
        """
        pass
        path = NestedDictAccessor.get_nested_value(
            data=task_input,
            key=ParsingTask.EXPECTED_KEY,
            default=ParsingTask.DEFAULT_DIRECTORY,
        )
        file_type = NestedDictAccessor.get_nested_value(
            data=self.task_properties,
            key="file_type",
            default=ParsingTask.DEFAULT_FILE_TYPE,
        )
        parsed_format = NestedDictAccessor.get_nested_value(
            data=self.task_properties,
            key="parsed_format",
            default=ParsingTask.DEFAULT_PARSED_FORMAT,
        )

        # Process the file based on its type,
        # Incase the output is required in markdown or base64 format, we need to convert the output to base64
        if file_type in ParsingTask.DOCUMENT_TYPE:
            intermediate_results = DocUtils._process_document(
                path,
                return_base64=parsed_format == ParsedFormat.BASE64
                or parsed_format == ParsedFormat.MARKDOWN,
            )
        elif file_type in ParsingTask.IMAGE_TYPE:
            intermediate_results = ImageUtils._process_image(
                path,
                return_base64=parsed_format == ParsedFormat.BASE64
                or parsed_format == ParsedFormat.MARKDOWN,
            )

        final_result = []
        # Convert the intermediate results to the required format
        if parsed_format == ParsedFormat.MARKDOWN:
            # Generate the final result
            for intermediate_result in intermediate_results:
                # Convert the image to markdown
                base64_image = (
                    ImageUtils._image_to_base64(intermediate_result)
                    if isinstance(intermediate_result, Image.Image)
                    else intermediate_result
                )
                # Generate the final result
                final_result.append(
                    self._trigger_generation(
                        task_input={
                            "image": base64_image,
                        }
                    )
                )
        else:
            final_result = intermediate_results

        # Return the final result
        return final_result

    @lru_cache
    def get_valid_completion_args(self) -> Set[str]:
        completion_params = inspect.signature(self.client.chat.completions.create).parameters
        # Return a set of valid argument names
        return set(completion_params.keys())

    def _trigger_generation(
        self,
        task_input: Dict[str, Any],
    ) -> Any:
        """
        Trigger the generation task.
        """
        # Get Prompt from Task Property
        generation_properties = copy.deepcopy(self.task_properties)
        generation_properties.pop("prompt", None)

        prompt = NestedDictAccessor.get_nested_value(
            self.task_properties,
            key="prompt",
            default=ParsingTask.DEFAULT_PROMPT,
        )

        # Format the Prompt
        formatted_prompt = GeneralUtils.populate_template(
            prompt,
            self.task_properties | task_input,  # Merge task_properties and task_input
        )

        # Format OpenAI Call
        content = [
            {
                "type": "text",
                "text": formatted_prompt,
            }
        ]

        img_keys = Base64ImageFinder.find_base64_encoded_keys(task_input=task_input)

        for img_key in img_keys:
            base64_image = NestedDictAccessor.get_nested_value(
                task_input,
                img_key,
                None,
            )
            if base64_image:
                image_arg = {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:image/jpeg;base64,{base64_image}",  # TODO(wizenheimer): add support for other image types
                        "detail": "low",  # TODO(wizenheimer): make this configurable
                    },
                }
                content.append(image_arg)

        # Create the messages list with the proper structure
        messages = [
            {
                "role": "user",
                "content": content,
            }
        ]

        # Update generation_property with the correct messages format
        generation_properties["messages"] = messages

        generation_properties["response_format"] = MarkdownModel

        # Filter out unnecessary properties form generation properties
        filtered_generation_property = {
            k: v for k, v in generation_properties.items() if k in self.get_valid_completion_args()
        }

        # Perform completion and handle errors
        response = self.client.beta.chat.completions.parse(**filtered_generation_property)

        # Return results
        if response.choices[0].message.refusal:
            logger.error(f"Request refused: {response.choices[0].message.refusal}")
            return None

        return response.choices[0].message.parsed.model_dump().get("markdown")  # type: ignore

    def _generate_references(
        self,
    ) -> List[Dict[str, Any]]:
        # Get the task properties
        directory = NestedDictAccessor.get_nested_value(
            data=self.task_properties,
            key="directory",
            default=ParsingTask.DEFAULT_DIRECTORY,
        )

        file_type = NestedDictAccessor.get_nested_value(
            data=self.task_properties,
            key="file_type",
            default=ParsingTask.DEFAULT_FILE_TYPE,
        )

        max_depth = NestedDictAccessor.get_nested_value(
            data=self.task_properties,
            key="max_depth",
            default=ParsingTask.DEFAULT_MAX_DEPTH,
        )

        parsed_list = FileUtils._parse_files(directory, file_type, max_depth)
        return [
            {
                ParsingTask.EXPECTED_KEY: path,
            }
            for path in parsed_list
        ]


class ImageUtils:
    @staticmethod
    def _process_image(
        image_path: str,
        return_base64: bool = True,
    ) -> Union[List[Image.Image], List[str]]:
        """
        Processes a single image file.
        """
        image = ImageUtils._read_image_file(image_path)
        return [ImageUtils._image_to_base64(image)] if return_base64 else [image]

    @staticmethod
    def _read_image_file(
        file_path: str,
    ):
        """
        Reads an image file from the given path.
        """
        return Image.open(file_path)

    @staticmethod
    def _image_to_base64(
        image,
        format: Optional[str] = None,
    ):
        """
        Converts an image file to a base64 string.
        """
        buffered = BytesIO()
        # Use the original image format if available and no format is specified
        if format is None:
            format = image.format or "PNG"
        image.save(buffered, format=format)
        return base64.b64encode(buffered.getvalue()).decode("utf-8")

    @staticmethod
    def _base64_to_image(
        base64_string: str,
    ):
        """
        Converts a base64 string to an image file.
        """
        image_data = base64.b64decode(
            base64_string,
        )
        return Image.open(BytesIO(image_data))


class AudioUtils:
    @staticmethod
    def _read_audio_file(
        file_path: str,
    ):
        """
        Reads an audio file from the given path.
        """
        return AudioSegment.from_file(
            file_path,
        )

    @staticmethod
    def _audio_to_base64(
        audio,
        format: str = "mp3",
    ):
        """
        Converts an audio file to a base64 string.
        """
        buffered = BytesIO()
        audio.export(
            buffered,
            format=format,
        )
        return base64.b64encode(buffered.getvalue()).decode("utf-8")

    @staticmethod
    def _base64_to_audio(
        base64_string: str,
    ):
        """
        Converts a base64 string to an audio file.
        """
        audio_data = base64.b64decode(
            base64_string,
        )
        return AudioSegment.from_file(BytesIO(audio_data))


class DocUtils:
    DPI = 300
    FMT = "png"
    SIZE = None
    THREAD_COUNT = 1
    USE_PDFTOCAIRO = False

    @staticmethod
    def _process_document(
        pdf_path: str,
        return_base64: bool = True,
        dpi: Optional[int] = None,
        fmt: Optional[str] = None,
        size: Optional[str] = None,
        thread_count: Optional[int] = None,
        use_pdftocairo: Optional[bool] = None,
    ) -> Union[List[Image.Image], List[str]]:
        """
        Processes a single PDF file.
        """

        options = {
            "dpi": dpi or DocUtils.DPI,
            "fmt": fmt or DocUtils.FMT,
            "size": size or DocUtils.SIZE,
            "thread_count": thread_count or DocUtils.THREAD_COUNT,
            "use_pdftocairo": use_pdftocairo or DocUtils.USE_PDFTOCAIRO,
        }

        try:
            images = convert_from_path(pdf_path, **options)

            if return_base64:
                return [ImageUtils._image_to_base64(img) for i, img in enumerate(images, start=1)]
            else:
                return images
        except Exception as err:
            warnings.warn(f"Error processing PDF: {err}")
            return []


class FileUtils:
    @staticmethod
    def _parse_files(
        directory: str,
        file_types: str,
        max_depth: int,
    ) -> List[str]:
        """
        Parse files in a directory and return a list of absolute file paths.
        """
        # Convert directory to absolute path
        directory = os.path.abspath(directory)

        # Parse files
        result = []

        def traverse(
            current_dir: str,
            current_depth: int,
        ):
            if current_depth > max_depth:
                return

            try:
                for entry in os.scandir(current_dir):
                    if entry.is_file():
                        _, ext = os.path.splitext(entry.name)
                        if ext[1:] == file_types:  # Remove the dot from extension
                            result.append(os.path.abspath(entry.path))
                    elif entry.is_dir():
                        traverse(entry.path, current_depth + 1)
            except PermissionError:
                warnings.warn(
                    f"Permission denied for directory: {current_dir}. Skipping this directory."
                )

        traverse(directory, 0)

        return result



from enum import Enum

import pytest
from cyyrus.models.task import Task  # type: ignore
from cyyrus.models.task_model import LargeLanguageModels  # type: ignore
from cyyrus.models.task_type import TASK_PROPERTIES, TaskType  # type: ignore
from cyyrus.models.types import (  # type: ignore
    DataType,
    ObjectProperty,
)
from pydantic import ValidationError


def create_task(task_type, properties):
    return Task(task_type=task_type, task_properties=properties)


def test_valid_task_creation():
    for task_type in TaskType:
        properties = {}
        for prop, (_, default) in TASK_PROPERTIES[task_type]["required"].items():
            if default is ...:
                properties[prop] = "test_value"
        task = create_task(task_type, properties)
        assert task.task_type == task_type
        assert all(prop in task.task_properties for prop in TASK_PROPERTIES[task_type]["required"])


def test_missing_required_property():
    with pytest.raises(ValidationError):
        create_task(TaskType.PARSING, {})


def test_invalid_property_type():
    with pytest.raises(ValidationError):
        create_task(TaskType.PARSING, {"directory": 123, "file_type": "txt"})


def test_all_task_types():
    for task_type in TaskType:
        properties = {}
        for prop, (_, default) in TASK_PROPERTIES[task_type]["required"].items():
            if default is ...:
                properties[prop] = "test_value"
        task = create_task(task_type, properties)
        assert task.task_type == task_type


def test_data_type():
    assert DataType.STRING == "string"
    assert isinstance(DataType.STRING, Enum)


def test_object_property():
    # Test with DataType
    prop1 = ObjectProperty(type=DataType.STRING)
    assert prop1.type == DataType.STRING

    # Test with invalid type
    with pytest.raises(ValidationError):
        ObjectProperty(type=123)  # type: ignore


def test_default_values():
    task = create_task(TaskType.PARSING, {"directory": "test", "file_type": "txt"})
    assert task.task_properties["max_depth"] == 5


def test_optional_property_override():
    task = create_task(TaskType.PARSING, {"directory": "test", "file_type": "txt", "max_depth": 10})
    assert task.task_properties["max_depth"] == 10


def test_extra_properties_ignored():
    task = create_task(
        TaskType.PARSING, {"directory": "test", "file_type": "txt", "extra_prop": "ignored"}
    )
    assert "extra_prop" not in task.task_properties


def test_enum_property():
    task = create_task(
        TaskType.GENERATION, {"prompt": "test", "model": LargeLanguageModels.GPT_4O_MINI}
    )
    assert isinstance(task.task_properties["model"], LargeLanguageModels)
    assert task.task_properties["model"] == LargeLanguageModels.GPT_4O_MINI


def test_enum_property_string_input():
    task = create_task(TaskType.GENERATION, {"prompt": "test", "model": "gpt-4o-mini"})
    assert isinstance(task.task_properties["model"], LargeLanguageModels)
    assert task.task_properties["model"] == LargeLanguageModels.GPT_4O_MINI


def test_missing_optional_property():
    task = create_task(TaskType.AUDIOEXTRACTION, {})
    assert "max_duration" in task.task_properties
    assert task.task_properties["max_duration"] == 60



class Messages:
    """
    Constants for messages
    """

    # ==============================
    # Schema Section validation errors
    # ==============================

    SCHEMA_FILE_COULD_NOT_BE_LOCATED = "Schema file could not be located. Please check the schema file path. Must be a valid file path or URL. For more details, see the https://cyyrus.com/docs/schema"

    SCHEMA_COULD_NOT_BE_PARSED = "Schema could not be parsed. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    SCHEMA_COULD_NOT_BE_VALIDATED = "Schema could not be validated. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    REQUIRED_SECTION_MISSING_IN_SCHEMA = "Required section missing in schema. For more details, see the https://cyyrus.com/docs/schema"

    # ==============================
    # Reference Section validation error
    # ==============================

    REFERENCE_TYPE_MISSING = "Reference type missing. Specify a valid reference type. For more details, see the https://cyyrus.com/docs/schema"

    REFERENCE_TYPE_NOT_SUPPORTED = (
        "Reference type not supported. For more details, see the https://cyyrus.com/docs/schema"
    )

    REFERENCE_PROPERTIES_MISSING = "Reference properties missing. Specify properties for the references. For more details, see the https://cyyrus.com/docs/schema"

    REFERENCE_PROPERTIES_INVALID = "Reference properties invalid. Specify valid properties for the references. For more details, see the https://cyyrus.com/docs/schema"

    # ==============================
    # Dataset Section validation error
    # ==============================

    SPLITS_DONT_ADD_UP = "Splits don't add up to 1. Please check the schema file. Would normalize the splits. For more details, see the https://cyyrus.com/docs/schema. Attempting to normalize the splits."

    SPLIT_VALUE_INVALID = "Split value is invalid. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema. Setting to absolute value."

    INVALID_TAG_TYPES = "Invalid tag types. Populating with default tags. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    INVALID_SAMPLING_STRATEGY = "Invalid sampling strategy. Populating with default sampling strategy. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    INVALID_NULL_HANDLING_STRATEGY = "Invalid null handling strategy. Populating with default null handling strategy of `include`. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    INVALID_REFERENCE_HANDLING_STRATEGY = "Invalid reference handling strategy. Populating with default reference handling strategy of `include`. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    INVALID_SHUFFLING_STRATEGY = "Invalid shuffling strategy. Populating with default shuffling strategy with seed `42`. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    # ==============================
    # Types Section validation error
    # ==============================

    INVALID_TYPE = "Invalid type. Specify a valid type. For more details, see the https://cyyrus.com/docs/schema"

    TYPE_PROPERTIES_MISSING_FOR_OBJECT_TYPE = "Type properties missing for object type. Specify `properties` for the object type. For more details, see the https://cyyrus.com/docs/schema"

    TYPE_ITEMS_MISSING_FOR_ARRAY_TYPE = "Type items missing for array type. Specify `items` for the array type. For more details, see the https://cyyrus.com/docs/schema"

    DUPLICATE_TYPE_NAME = "Duplicate type name found. Type names must be unique. For more details, see the https://cyyrus.com/docs/schema"

    MAXIMUM_DEPTH_EXCEEDED = "Maximum depth exceeded. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema."

    REFERENCE_TYPE_NOT_FOUND = "Reference type not found. Specify a valid reference type. Defaulting to `string`. For more details, see the https://cyyrus.com/docs/schema"

    # Task Section validation error

    TASK_SPECIFICATION_MISSING = "Task specification missing. Specify either of `task_id` or `task_type`. For more details, see the https://cyyrus.com/docs/schema"

    TASK_PROPERTIES_UNSPECIFIED = "Task properties unspecified. Specify `task_properties` for the task. For more details, see the https://cyyrus.com/docs/schema"

    REFERENCED_TASK_ID_NOT_FOUND = "Referenced task_id not found. Specify a valid task_id. For more details, see the https://cyyrus.com/docs/schema"

    TASK_INPUTS_CYCLIC_DEPENDENCY = "Task inputs have cyclic dependency. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    # Column Section validation error

    COLUMN_TYPE_NOT_FOUND = "Column type not found. Specify a valid column type. Defaulting to `string`. For more details, see the https://cyyrus.com/docs/schema"

    COLUMN_TASK_ID_NOT_FOUND = "Column task_id not found. Specify a valid task_id. For more details, see the https://cyyrus.com/docs/schema"

    COLUMNN_ID_NOT_FOUND = "Column ID not found. Specify a valid column ID. For more details, see the https://cyyrus.com/docs/schema"

    DUPLICATE_COLUMN_NAME = "Duplicate column name found. Column names must be unique across `reference` and `columns` sections. For more details, see the https://cyyrus.com/docs/schema"

    # ==============================
    #  Composer Section validation error
    # ==============================
    KEY_COLUMN_NOT_FOUND = "Key column not found. Specify a valid key column. For more details, see the https://cyyrus.com/docs/schema"

    INVALID_KEY_COLUMN = "Invalid key column. Specify a valid key column. For more details, see the https://cyyrus.com/docs/schema"

    ALL_ROWS_EXCLUDED_DUE_TO_NAN = "All rows excluded due to NaN. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    REQUIRED_COLUMN_MISSING = "Required column missing. Creating them with empty values. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    NON_UNIQUE_COLUMN_VALUES = "Removed non-unique column values. Please check the schema file. For more details, see the https://cyyrus.com/docs/schema"

    DATASET_TOO_SMALL_FOR_SPLIT = "Dataset too small for split. Using entire dataset as train set. For more details, see the https://cyyrus.com/docs/schema"

    RE_ADJUSTING_SPLIT = "Re-adjusting split values. Adjusted split sizes would be unstable or exceed dataset size. For more details, see the https://cyyrus.com/docs/schema"



import random


class FunnyBones:
    funny_words = [
        "aardvark",
        "abacus",
        "abundance",
        "ache",
        "acupuncture",
        "airbrush",
        "alien",
        "anagram",
        "angle",
        "amazing",
        "ankle",
        "alphabet",
        "antenna",
        "aqua",
        "asphalt",
        "bacon",
        "banana",
        "bangles",
        "banjo",
        "bankrupt",
        "bar",
        "barracuda",
        "basket",
        "beluga",
        "binder",
        "birthday",
        "bisect",
        "blizzard",
        "blunderbuss",
        "boa",
        "bog",
        "bounce",
        "broomstick",
        "brought",
        "bubble",
        "budgie",
        "bug",
        "bug-a-boo",
        "bugger",
        "buff",
        "burst",
        "butter",
        "buzz",
        "cabana",
        "cake",
        "calculator",
        "camera",
        "candle",
        "carnival",
        "carpet",
        "casino",
        "cashew",
        "catfish",
        "ceiling",
        "celery",
        "chalet",
        "chalk",
        "chart",
        "cheddar",
        "chesterfield",
        "chicken",
        "chinchill",
        "chit-chat",
        "chocolate",
        "chowder",
        "coal",
        "compass",
        "compress",
        "computer",
        "conduct",
        "contents",
        "cookie",
        "copper",
        "corduroy",
        "cow",
        "cracker",
        "crackle",
        "croissant",
        "cube",
        "cupcake",
        "curly",
        "curtain",
        "cushion",
        "cuticle",
        "daffodil",
        "delicious",
        "dictionary",
        "dimple",
        "ding-a-ling",
        "disk",
        "disco duck",
        "dodo",
        "dolphin",
        "dong",
        "donuts",
        "dork",
        "dracula",
        "duct tape",
        "effigy",
        "egad",
        "elastic",
        "elephant",
        "encasement",
        "erosion",
        "eyelash",
        "fabulous",
        "fantastic",
        "feather",
        "felafel",
        "fetish",
        "financial",
        "finger",
        "finite",
        "fish",
        "fizzle",
        "fizzy",
        "flame",
        "flash",
        "flavour",
        "flick",
        "flock",
        "flour",
        "flower",
        "foamy",
        "foot",
        "fork",
        "fritter",
        "fudge",
        "fungus",
        "funny",
        "fuse",
        "fusion",
        "fuzzy",
        "garlic",
        "gelatin",
        "gelato",
        "ghetto",
        "glebe",
        "glitter",
        "glossy",
        "groceries",
        "goulashes",
        "guacamole",
        "gumdrop",
        "haberdashery",
        "hamster",
        "happy",
        "highlight",
        "hippopotamus",
        "hobbit",
        "hold",
        "hooligan",
        "hydrant",
        "icicles",
        "idiot",
        "implode",
        "implosion",
        "indeed",
        "issue",
        "itchy",
        "jell-o",
        "jewel",
        "jump",
        "kabob",
        "kasai",
        "kite",
        "kiwi",
        "ketchup",
        "knob",
        "laces",
        "lacy",
        "laughter",
        "laundry",
        "leaflet",
        "legacy",
        "leprechaun",
        "lollypop",
        "lumberjack",
        "macadamia",
        "magenta",
        "magic",
        "magnanimous",
        "mango",
        "margarine",
        "massimo",
        "mechanical",
        "medicine",
        "meh",
        "melon",
        "meow",
        "mesh",
        "metric",
        "microphone",
        "minnow",
        "mitten",
        "mozzarella",
        "muck",
        "mumble",
        "mushy",
        "mustache",
        "nanimo",
        "noodle",
        "nostril",
        "nuggets",
        "oatmeal",
        "oboe",
        "o'clock",
        "octopus",
        "odour",
        "ointment",
        "olive",
        "optic",
        "overhead",
        "ox",
        "oxen",
        "pajamas",
        "pancake",
        "pansy",
        "paper",
        "paprika",
        "parmesan",
        "pasta",
        "pattern",
        "pecan",
        "peek-a-boo",
        "pen",
        "pepper",
        "pepperoni",
        "peppermint",
        "perfume",
        "periwinkle",
        "photograph",
        "pie",
        "pierce",
        "pillow",
        "pimple",
        "pineapple",
        "pistachio",
        "plush",
        "polish",
        "pompom",
        "poodle",
        "pop",
        "popsicle",
        "prism",
        "prospector",
        "prosper",
        "pudding",
        "puppet",
        "puzzle",
        "queer",
        "query",
        "radish",
        "rainbow",
        "ribbon",
        "rotate",
        "salami",
        "sandwich",
        "saturday",
        "saturn",
        "saxophone",
        "scissors",
        "scooter",
        "scrabbleship",
        "scrunchie",
        "scuffle",
        "shadow",
        "sickish",
        "silicone",
        "slippery",
        "smash",
        "smooch",
        "snap",
        "snooker",
        "socks",
        "soya",
        "spaghett",
        "sparkle",
        "spatula",
        "spiral",
        "splurge",
        "spoon",
        "sprinkle",
        "square",
        "squiggle",
        "squirrel",
        "statistics",
        "stuffing",
        "sticky",
        "sugar",
        "sunshine",
        "super",
        "swirl",
        "taffy",
        "tangy",
        "tape",
        "tat",
        "teepee",
        "telephone",
        "television",
        "thinkable",
        "tip",
        "tofu",
        "toga",
        "trestle",
        "tulip",
        "turnip",
        "turtle",
        "tusks",
        "ultimate",
        "unicycle",
        "unique",
        "uranus",
        "vegetable",
        "waddle",
        "waffle",
        "wallet",
        "walnut",
        "wagon",
        "window",
        "whatever",
        "whimsical",
        "wobbly",
        "yellow",
        "zap",
        "zebra",
        "zigzag",
        "zip",
    ]

    @staticmethod
    def suggest():
        adjective = random.choice(FunnyBones.funny_words)
        noun = random.choice(FunnyBones.funny_words)
        while adjective == noun:  # Ensure adjective and noun are different
            noun = random.choice(FunnyBones.funny_words)
        return f"{adjective}_{noun}"



from typing import List

from pydantic import BaseModel, Field


class Column(BaseModel):
    task_id: str = Field(
        ...,
        description="ID of the task associated with the column",
    )
    task_input: List[str] = Field(
        default=[],
        description="Input to the task",
    )
    description: str = Field(
        default="",
        description="Description of the column",
    )



import logging
from functools import lru_cache
from logging import handlers
from pathlib import Path
from typing import Dict, Optional

import colorlog
from pythonjsonlogger import jsonlogger
from tqdm import tqdm

# Default silenced loggers
DEFAULT_SILENCED_LOGGERS = {
    "urllib3": logging.WARNING,
    "requests": logging.WARNING,
    "matplotlib": logging.WARNING,
    "paramiko": logging.WARNING,
    "boto3": logging.WARNING,
    "botocore": logging.WARNING,
    "nose": logging.WARNING,
}


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


def setup_logging(
    log_level=logging.INFO,
    log_file=None,
    for_human=True,
    log_dir: Optional[Path] = None,
    silenced_loggers: Optional[Dict[str, int]] = None,
):
    """
    Set up logging for the application.
    """
    # Use provided log_dir if given, otherwise use default
    if log_dir is None:
        log_dir = Path(__file__).parent.parent.parent.parent.parent / "logs"

    # Create logs directory if it doesn't exist
    log_dir.mkdir(exist_ok=True)

    # Set up root logger
    root_logger = logging.getLogger("cyyrus")
    root_logger.setLevel(log_level)

    # Create formatters
    if for_human:
        console_formatter = colorlog.ColoredFormatter(
            "%(log_color)s%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )
        file_formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(message)s",
        )
    else:
        console_formatter = jsonlogger.JsonFormatter(
            "%(asctime)s %(name)s %(levelname)s %(message)s %(pathname)s %(lineno)d",
        )
        file_formatter = jsonlogger.JsonFormatter(
            "%(asctime)s %(name)s %(levelname)s %(message)s %(pathname)s %(lineno)d",
        )

    # TQDM handler
    console_handler = TqdmLoggingHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(console_formatter)
    root_logger.addHandler(console_handler)

    # File handler (optional)
    if log_file:
        file_handler = handlers.RotatingFileHandler(
            log_dir / log_file,
            maxBytes=10 * 1024 * 1024,  # 10MB
            backupCount=5,
        )
        file_handler.setLevel(log_level)
        file_handler.setFormatter(file_formatter)
        root_logger.addHandler(file_handler)

    # Silence chatty libraries
    silenced_loggers = silenced_loggers or DEFAULT_SILENCED_LOGGERS
    for logger_name, level in silenced_loggers.items():
        logging.getLogger(logger_name).setLevel(level)

    return root_logger


@lru_cache()
def get_logger(
    name: Optional[str] = None,
) -> logging.Logger:
    """
    Retrieves a logger with the given name, or the root logger if no name is given.

    Args:
        name: The name of the logger to retrieve.

    Returns:
        The logger with the given name, or the root logger if no name is given.

    Example:
        Basic Usage of `get_logger`
        ```python
        from cyyrus.utils.logging_config import get_logger

        logger = get_logger("cyyrus.test")
        logger.info("This is a test") # Output: cyyrus.test: This is a test

        debug_logger = get_logger("cyyrus.debug")
        debug_logger.debug("Debug message")
        ```
    """
    parent_logger = logging.getLogger("cyyrus")

    if name:
        # Append the name if given but allow explicit full names e.g. "cyyrus.test"
        # should not become "cyyrus.cyyrus.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    return logger


def deprecated(
    message: str,
    version: str,
):
    """
    Decorator to mark a function as deprecated.

    Args:
        message (str): The deprecation message.
        version (str): The version in which the function is deprecated.

    Returns:
        function: The decorated function.

    Example:
        @deprecated("This function is deprecated", "1.0")
        def my_function():
            pass
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            get_logger(__file__).warning(
                f"WARNING: {func.__name__} is deprecated as of version {version}. {message}".strip(),
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


# if __name__ == "__main__":
#     setup_logging(
#         log_level=logging.DEBUG,
#         log_file="cyyrus.log",
#         for_human=True,
#     )
#     logger = get_logger("cyyrus.test")
#     logger.debug("This is a debug message")
#     logger.info("This is an info message")
#     logger.warning("This is a warning message")
#     logger.error("This is an error message")
#     logger.critical("This is a critical message")

#     @deprecated("Use new_function instead", "1.0")
#     def old_function():
#         pass

#     old_function()



from typing import Any, Dict

from pydantic import BaseModel
from pydantic.config import ConfigDict
from pydantic.functional_validators import model_validator

from cyyrus.models.task_type import TASK_PROPERTIES, TaskType
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class Task(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    task_type: TaskType
    task_properties: Dict[str, Any]

    @model_validator(mode="after")
    def validate_task_properties(self):
        """
        Validate the task properties
        """
        logger.debug(f"Validating task properties for {self.task_type}")
        required_props, optional_props = TaskPropertyUtils.get_task_property(
            self.task_type,
        )

        all_props = {
            **required_props,
            **optional_props,
        }

        for prop, (prop_type, default_value) in all_props.items():
            if prop in self.task_properties:
                try:
                    self.task_properties[prop] = prop_type(
                        self.task_properties[prop],
                    )
                except (ValueError, TypeError) as e:
                    raise ValueError(f"Invalid value for {prop}: {str(e)}")
            elif prop in required_props and default_value is ...:
                raise ValueError(f"Missing required property {prop} for {self.task_type}")
            elif default_value is not ...:
                self.task_properties[prop] = default_value

        logger.debug(f"Task properties {list(self.task_properties.keys())[:2]}... validated")
        return self


class TaskPropertyUtils:
    @staticmethod
    def get_task_property(
        task_type: TaskType,
    ) -> Any:
        """
        Get the required and optional properties for a given task type
        """
        properties = TASK_PROPERTIES.get(
            task_type,
            {},
        )
        required_props = properties.get(
            "required",
            {},
        )
        optional_props = properties.get(
            "optional",
            {},
        )

        logger.debug(f"Fetched Task properties for {task_type}")
        return required_props, optional_props



import functools
import logging
from typing import Any, Callable, Optional, Type, Union

from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


def error_handler(
    exceptions: Union[Type[Exception], tuple[Type[Exception], ...]] = Exception,
    handler: Optional[Callable[[Exception], Any]] = None,
    logger: Optional[logging.Logger] = logger,
    retries: int = 0,
    default_return: Any = None,
):
    """
    A decorator factory for custom error handling and returning a default value incase of retires get exhausted.

    Args:
        exceptions: Exception or tuple of exceptions to catch.
        handler: Function to handle the caught exception.
        logger: Logger object to use for logging errors.
        retries: Number of times to retry the function before giving up.
        default_return: Value to return if all retries fail.

    Returns:
        Decorator function for error handling.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(retries + 1):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    if logger:
                        # TODO: explore using logger.exception() instead
                        logger.error(
                            f"Error in {func.__name__}. Attempt {attempt + 1} of {retries + 1}"
                        )

                    if handler:
                        handler_result = handler(e)
                        # If the handler returns a value, use it as the final result
                        if handler_result is not None:
                            return handler_result

                    # If we've reached the last attempt, return the default
                    if attempt == retries:
                        return default_return

            # This line should never be reached, but it's here for completeness
            return default_return

        return wrapper

    return decorator



from typing import Dict, List

import pytest
from cyyrus.models.spec import level_order_traversal  # type: ignore


@pytest.mark.parametrize(
    "dependencies, expected_output",
    [
        (
            {
                "document_id": ["invoice"],
                "document_info": ["invoice"],
                "invoice_items": ["invoice"],
                "invoice_questions": ["invoice_items"],
            },
            [["invoice"], ["document_id", "document_info", "invoice_items"], ["invoice_questions"]],
        ),
        ({"A": ["B", "C"], "B": ["D"], "C": ["D"], "D": []}, [["D"], ["B", "C"], ["A"]]),
        ({"A": [], "B": ["A"], "C": ["A"], "D": ["B", "C"]}, [["A"], ["B", "C"], ["D"]]),
        ({}, []),
        ({"A": []}, [["A"]]),
        ({"A": ["B"], "B": ["A"]}, []),  # Cyclic dependency should result in empty list
    ],
)
def test_level_order_traversal(
    dependencies: Dict[str, List[str]], expected_output: List[List[str]]
):
    assert list(level_order_traversal(dependencies)) == expected_output


def test_missing_dependency():
    dependencies = {"A": ["B"], "C": ["B"]}
    result = list(level_order_traversal(dependencies))
    assert "B" in result[0], "B should be in the first level as it's not in the dependencies dict"
    assert set(result[1]) == {"A", "C"}, "A and C should be in the second level"


def test_all_nodes_included():
    dependencies = {"A": ["B", "C"], "D": ["E"]}
    result = list(level_order_traversal(dependencies))
    all_nodes = set(sum(result, []))  # Flatten the result
    assert all_nodes == {"A", "B", "C", "D", "E"}, "All nodes should be included in the result"


def test_order_within_levels():
    dependencies = {"A": ["D"], "B": ["D"], "C": ["D"], "D": []}
    result = list(level_order_traversal(dependencies))
    assert result[0] == ["D"], "D should be in the first level"
    assert set(result[1]) == {"A", "B", "C"}, "A, B, C should be in the second level"
    assert len(result) == 2, "There should be only two levels"



import base64
import difflib
import re
from typing import Any, Dict, List, TypeVar, Union

from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)

T = TypeVar("T")


class GeneralUtils:
    @staticmethod
    def extend_or_append(
        lst: List[T],
        item: Union[T, List[T]],
    ) -> List[T]:
        """
        Extends the list if item is a list, or appends item if it's a single element.

        :param lst: The list to be modified
        :param item: The item or list of items to be added
        :return: The modified list
        """
        if isinstance(item, list):
            lst.extend(item)
        else:
            lst.append(item)
        return lst

    @staticmethod
    def flatten_dict(
        d: Dict[str, Any],
        parent_key: str = "",
        sep: str = "_",
        max_depth: int = 5,
    ) -> Dict[str, Any]:
        """
        Attempts to flatten the dictionary to a single level, by concatenating the keys with the separator. Upto the specified max_depth.
        """
        logger.debug(f"Attempting to flatten output with max_depth {max_depth} ...")
        items = []
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict) and max_depth > 1:
                items.extend(
                    GeneralUtils.flatten_dict(
                        v,
                        new_key,
                        sep,
                        max_depth - 1,
                    ).items()
                )
            else:
                items.append((new_key, v))
        return dict(items)

    @staticmethod
    def populate_template(
        template: str,
        data: Dict[str, Any],
    ) -> str:
        def get_nested_value(
            keys: List[str],
            nested_data: Dict[str, Any],
        ) -> Any:
            """Recursively get value from nested dictionary."""
            for key in keys:
                if isinstance(nested_data, dict) and key in nested_data:
                    nested_data = nested_data[key]
                else:
                    return None
            return nested_data

        def replace_placeholder(
            match,
        ):
            placeholder = match.group(1).strip()
            keys = placeholder.split(".")
            value = get_nested_value(keys, data)
            return str(value) if value is not None else f"{{{placeholder}}}"

        pattern = r"\{([^}]+)\}"
        return re.sub(pattern, replace_placeholder, template)


class NestedDictAccessor:
    MAX_DEPTH: int = 4

    @staticmethod
    def _find_closest_key(
        keys: List[str],
        target: str,
    ) -> str:
        """
        Find the closest matching key from the list of keys.
        """
        if not keys:
            return ""
        return max(
            keys,
            key=lambda k: difflib.SequenceMatcher(None, k, target).ratio(),
        )

    @staticmethod
    def get_nested_value(
        data: Dict[str, Any],
        key: str,
        default: Any = None,
    ) -> Any:
        """
        Retrieve a value from a nested dictionary using a dot-separated key string.
        If the exact key is not found, it attempts to find the closest matching key at each level.

        :param data: The input dictionary
        :param key: Dot-separated string representing the nested key
        :param default: Default value to return if key is not found
        :return: The value found or the default
        """

        def _get_value(
            current_data: Dict[str, Any],
            keys: List[str],
            current_depth: int = 0,
        ) -> Any:
            if (
                current_depth >= NestedDictAccessor.MAX_DEPTH
                or not keys
                or not isinstance(current_data, dict)
            ):
                return current_data

            current_key = keys[0]
            if current_key in current_data:
                return _get_value(current_data[current_key], keys[1:], current_depth + 1)
            else:
                logger.debug(f"Key '{current_key}' not found, attempting to find closest match...")
                closest_key = NestedDictAccessor._find_closest_key(
                    list(current_data.keys()),
                    current_key,
                )
                if closest_key:
                    logger.debug(f"Closest key found: '{closest_key}'")
                    return _get_value(
                        current_data[closest_key],
                        keys[1:],
                        current_depth + 1,
                    )
                else:
                    logger.debug(f"No suitable key found for '{current_key}'")
                    return default

        key_parts = key.split(".")
        result = _get_value(data, key_parts)
        return result if result is not None else default


class Base64ImageFinder:
    MAX_DEPTH: int = 5

    @staticmethod
    def is_base64_image(
        value: str,
    ) -> bool:
        """
        Check if a string is a base64 encoded image.
        """
        logger.debug("Checking if string is a base64 encoded image ...")
        try:
            # Attempt to decode the string
            decoded = base64.b64decode(value)
            # Check for common image headers (you might want to expand this list)
            return decoded.startswith(
                (
                    b"\xFF\xD8\xFF",  # JPEG
                    b"\x89PNG\r\n\x1a\n",  # PNG
                    b"GIF87a",  # GIF
                    b"GIF89a",  # GIF
                    b"RIFF",  # WEBP
                )
            )
        except Exception as e:
            logger.debug(f"Error decoding base64: {e}")
            return False

    @staticmethod
    def find_base64_encoded_keys(
        task_input: Dict[str, Any],
    ) -> List[str]:
        """
        Find base64 encoded images inside input up to 5 levels of nesting.
        """
        logger.debug("Attempting to find multimodal keys in task input ...")

        def search_nested(
            data: Dict[str, Any], current_path: str = "", current_depth: int = 0
        ) -> List[str]:
            keys = []
            if current_depth >= Base64ImageFinder.MAX_DEPTH:
                return keys

            for key, value in data.items():
                new_path = f"{current_path}.{key}" if current_path else key

                if isinstance(value, str) and Base64ImageFinder.is_base64_image(value):
                    keys.append(new_path)
                elif isinstance(value, dict):
                    keys.extend(search_nested(value, new_path, current_depth + 1))
                elif isinstance(value, list):
                    for i, item in enumerate(value):
                        if isinstance(item, dict):
                            keys.extend(search_nested(item, f"{new_path}[{i}]", current_depth + 1))
                        elif isinstance(item, str) and Base64ImageFinder.is_base64_image(item):
                            keys.append(f"{new_path}[{i}]")

            return keys

        return search_nested(task_input)



import pandas as pd
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree

from cyyrus.models.dataset import Dataset
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class Visualizer:

    @staticmethod
    def display_dataframe_properties(
        df: pd.DataFrame,
    ):
        """
        Display a concise overview of DataFrame properties and optionally visualize data distributions.

        Args:
        df (pd.DataFrame): The input DataFrame.
        logger: Optional logger object for additional logging.
        visualize (bool): Whether to create and display visualizations. Default is True.
        max_categories (int): Maximum number of categories to display for categorical columns. Default is 5.
        """

        console = Console()

        # Basic DataFrame info
        console.print(Panel.fit("[bold blue]DataFrame Overview", style="cyan"))
        console.print(
            f"Shape: [green]{df.shape[0]}[/green] rows, [green]{df.shape[1]}[/green] columns"
        )
        console.print(
            f"Memory usage: [yellow]{df.memory_usage(deep=True).sum() / 1024**2:.2f} MB[/yellow]"
        )

        # Column info
        column_table = Table(title="Column Information", box=box.ROUNDED)
        column_table.add_column("Column Name", style="cyan", no_wrap=True)
        column_table.add_column("Data Type", style="magenta")
        column_table.add_column("Non-Null Count", justify="right", style="green")
        column_table.add_column("Null Count", justify="right", style="red")

        for col in df.columns:
            non_null_count = df[col].count()
            null_count = df[col].isnull().sum()
            column_table.add_row(
                col,
                str(df[col].dtype),
                str(non_null_count),
                str(null_count),
            )

        console.print(column_table)

        # Data types overview
        dtype_tree = Tree("[bold]Data Types Overview")
        for dtype in df.dtypes.value_counts().index:
            dtype_cols = df.select_dtypes(include=[dtype]).columns
            branch = dtype_tree.add(f"[yellow]{dtype}[/yellow]: {len(dtype_cols)} columns")
            for col in dtype_cols:
                branch.add(f"[cyan]{col}[/cyan]")

        console.print(dtype_tree)

    @staticmethod
    def display_dataset_properties(dataset: Dataset):
        """
        Display properties of a Dataset object in a visually appealing way.

        Args:
        dataset (Dataset): The Dataset object to display.
        """
        console = Console()

        # Dataset Metadata
        console.print(Panel.fit("[bold blue]Dataset Metadata", style="cyan"))
        metadata_table = Table(box=box.ROUNDED)
        metadata_table.add_column("Property", style="cyan")
        metadata_table.add_column("Value", style="yellow")

        metadata_table.add_row("Name", dataset.metadata.name)
        metadata_table.add_row("Description", dataset.metadata.description)
        metadata_table.add_row("Tags", ", ".join(dataset.metadata.tags))
        metadata_table.add_row("License", dataset.metadata.license)
        metadata_table.add_row("Languages", ", ".join(dataset.metadata.languages))

        console.print(metadata_table)

        # Dataset Shuffle
        console.print(Panel.fit("[bold blue]Dataset Shuffle", style="cyan"))
        console.print(f"Seed: [yellow]{dataset.shuffle.seed}[/yellow]")

        # Dataset Splits
        console.print(Panel.fit("[bold blue]Dataset Splits", style="cyan"))
        splits_table = Table(box=box.ROUNDED)
        splits_table.add_column("Split", style="cyan")
        splits_table.add_column("Value", style="yellow")

        splits_table.add_row("Train", str(dataset.splits.train))
        splits_table.add_row("Test", str(dataset.splits.test))
        splits_table.add_row("Seed", str(dataset.splits.seed))

        console.print(splits_table)

        # Dataset Attributes
        console.print(Panel.fit("[bold blue]Dataset Attributes", style="cyan"))
        attributes_tree = Tree("[bold]Attributes")

        attributes_tree.add(
            f"Required Columns: [yellow]{', '.join(dataset.attributes.required_columns) or 'None'}[/yellow]"
        )
        attributes_tree.add(
            f"Unique Columns: [yellow]{', '.join(dataset.attributes.unique_columns) or 'None'}[/yellow]"
        )
        attributes_tree.add(f"Nulls: [yellow]{dataset.attributes.nulls}[/yellow]")
        attributes_tree.add(
            f"Flatten Columns: [yellow]{', '.join(dataset.attributes.flatten_columns) or 'None'}[/yellow]"
        )
        attributes_tree.add(
            f"Exclude Columns: [yellow]{', '.join(dataset.attributes.exclude_columns) or 'None'}[/yellow]"
        )

        console.print(attributes_tree)



import pandas as pd
import pytest
from cyyrus.composer.dataframe import (  # type: ignore
    ensure_required_columns,
    ensure_unique_columns,
    handle_nulls,
    normalize_split_sizes,
    split_dataset,
)
from datasets import Dataset


@pytest.fixture
def sample_df():
    return pd.DataFrame(
        {
            "A": [1, 2, 3, None, 5],
            "B": ["a", "b", "c", "d", "e"],
            "C": [1.1, 2.2, 3.3, 4.4, 5.5],
        }
    )


def test_handle_nulls(sample_df):
    # Test include nulls
    result = handle_nulls(sample_df, "include")
    assert len(result) == len(sample_df)

    # Test exclude nulls
    result = handle_nulls(sample_df, "exclude")
    assert len(result) == len(sample_df) - 1


def test_ensure_required_columns(sample_df):
    with pytest.warns():
        result = ensure_required_columns(sample_df, ["A", "B", "D"])
    assert "D" in result.columns
    assert result["D"].isnull().all()


def test_ensure_unique_columns(sample_df):
    df_with_duplicates = pd.concat([sample_df, sample_df.iloc[[0]]], ignore_index=True)
    with pytest.warns():
        result = ensure_unique_columns(df_with_duplicates, ["B"])
    assert len(result) == len(sample_df)


def test_normalize_split_sizes():
    # Test normal case
    train, test = normalize_split_sizes(0.8, 0.2)
    assert train == 0.8 and test == 0.2

    # Test case where sum != 1
    with pytest.warns():
        train, test = normalize_split_sizes(0.8, 0.3)
    assert pytest.approx(train + test) == 1


def test_split_dataset():
    dataset = Dataset.from_pandas(pd.DataFrame({"A": range(100)}))
    train, test = split_dataset(dataset, 0.8, 0.2, 42)
    assert len(train) == 80
    assert len(test) == 20

    # Test with extreme values
    train, test = split_dataset(dataset, 0.999, 0.001, 42)
    assert len(train) >= 1
    assert len(test) >= 1



from typing import Any, Dict, List

from cyyrus.models.task_type import TaskType
from cyyrus.tasks.base import BaseTask
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)


class DefaultTask(BaseTask):
    # The task ID is used to identify the task type
    TASK_ID = TaskType.DEFAULT

    # This flag indicates whether the task supports reference-free execution
    SUPPORTS_REFERENCE_FREE_EXECUTION = True

    # The execute method is the main method that needs to be implemented by the task
    def execute(
        self,
        task_input: Dict[str, Any],
    ) -> Any:
        """
        Perform the task execution
        """
        logger.debug(f"Executing default task with {self.TASK_ID} ...")
        return None

    # Incase the task supports reference-free execution, the task should implement the _generate_references method as well
    def _generate_references(
        self,
    ) -> List[Dict[str, Any]]:
        """
        Attempt to generate the reference data for the task, using the task properties
        """
        logger.debug(f"Generating references for task {self.TASK_ID} ...")
        return []



from typing import Any, Iterable, Optional, TypeVar

T = TypeVar("T")


def conditional_tqdm(
    iterable: Iterable[T],
    use_tqdm: bool = True,
    desc: Optional[str] = None,
    total: Optional[int] = None,
    **kwargs: Any,
):
    """
    A wrapper around tqdm that allows conditional use of the progress bar.

    Args:
        iterable (Iterable[T]): The iterable to wrap with tqdm.
        use_tqdm (bool): Whether to use tqdm or not. Defaults to True.
        desc (str, optional): Description to use for the progress bar.
        total (int, optional): The total number of iterations. If None, len(iterable) is used if possible.
        **kwargs: Additional keyword arguments to pass to tqdm.

    Returns:
        Iterator[T]: The wrapped iterable, either with tqdm progress bar or the original iterable.

    Example:
        ```python
        for item in tqdm_wrapper(items, use_tqdm=True, desc="Processing"):
            process(item)
        ```
    """
    if use_tqdm:
        from tqdm import tqdm

        return tqdm(iterable, desc=desc, total=total, **kwargs)
    else:
        return iter(iterable)



import pytest
from cyyrus.models.column import Column  # type: ignore
from pydantic import ValidationError


def test_column():
    # Test valid Column
    valid_column = Column(column_type="string", task_id="task1")
    assert valid_column.column_type == "string"
    assert valid_column.description == ""
    assert valid_column.task_id == "task1"
    assert valid_column.task_input == []

    # Test Column with all fields
    full_column = Column(
        column_type="integer",
        description="An integer column",
        task_id="task2",
        task_input=[
            "param1",
            "value1",
        ],
    )
    assert full_column.column_type == "integer"
    assert full_column.description == "An integer column"
    assert full_column.task_id == "task2"
    assert full_column.task_input == [
        "param1",
        "value1",
    ]

    # Test invalid Column (missing required field)
    with pytest.raises(ValidationError):
        Column(description="Invalid column")  # type: ignore



import pytest
from cyyrus.models.dataset import (  # type: ignore
    Dataset,
    DatasetAttributes,
    DatasetMetadata,
    DatasetShuffle,
    DatasetSplits,
)
from pydantic import ValidationError


def test_dataset_metadata():
    # Test default values
    default_metadata = DatasetMetadata()
    assert default_metadata.name == "Dataset"
    assert default_metadata.description == "Dataset generated using cyyrus"
    assert default_metadata.tags == ["dataset"]
    assert default_metadata.license == "MIT"

    # Test custom values
    custom_metadata = DatasetMetadata(
        name="Custom Dataset",
        description="A custom dataset",
        tags=["custom", "test"],
        license="Apache-2.0",
    )
    assert custom_metadata.name == "Custom Dataset"
    assert custom_metadata.description == "A custom dataset"
    assert custom_metadata.tags == ["custom", "test"]
    assert custom_metadata.license == "Apache-2.0"


def test_dataset_shuffle():
    # Test default values
    default_shuffle = DatasetShuffle()
    assert default_shuffle.seed == 42

    # Test custom values
    custom_shuffle = DatasetShuffle(seed=123, buffer_size=500)
    assert custom_shuffle.seed == 123

    # Test invalid seed
    with pytest.raises(ValidationError):
        DatasetShuffle(seed="awesome")  # type: ignore


def test_dataset_splits():
    # Test default values
    with pytest.warns():
        default_splits = DatasetSplits()
    assert default_splits.train == 0.8
    assert default_splits.test == 0.2

    # Test custom values
    custom_splits = DatasetSplits(train=0.5, test=0.5)
    assert custom_splits.train == 0.5
    assert custom_splits.test == 0.5

    # Test invalid values
    with pytest.raises(ValidationError):
        DatasetSplits(train=1.1)  # type: ignore
    with pytest.raises(ValidationError):
        DatasetSplits(train=-0.1)  # type: ignore

        # Test splits normalization
    normalized_splits = DatasetSplits(train=0.8, test=0.1)
    assert normalized_splits.train + normalized_splits.test == pytest.approx(1.0)  # type: ignore


def test_dataset_attributes():
    # Test default values
    default_attrs = DatasetAttributes()
    assert default_attrs.required_columns == []
    assert default_attrs.unique_columns == []
    assert default_attrs.nulls == "include"

    # Test custom values
    custom_attrs = DatasetAttributes(
        required_columns=["col1", "col2"],
        unique_columns=["col3"],
        nulls="exclude",
    )
    assert custom_attrs.required_columns == ["col1", "col2"]
    assert custom_attrs.unique_columns == ["col3"]
    assert custom_attrs.nulls == "exclude"

    # Test invalid values
    with pytest.raises(ValidationError):
        DatasetAttributes(nulls="invalid")


def test_dataset():
    # Test default values
    with pytest.warns():
        default_dataset = Dataset()
    assert isinstance(default_dataset.metadata, DatasetMetadata)
    assert isinstance(default_dataset.shuffle, DatasetShuffle)
    assert isinstance(default_dataset.splits, DatasetSplits)
    assert isinstance(default_dataset.attributes, DatasetAttributes)

    # Test custom values
    custom_dataset = Dataset(
        metadata=DatasetMetadata(name="Custom Dataset"),
        shuffle=DatasetShuffle(seed=123),
        splits=DatasetSplits(train=0.7, test=0.3, seed=42),
        attributes=DatasetAttributes(required_columns=["col1"]),
    )
    assert custom_dataset.metadata.name == "Custom Dataset"
    assert custom_dataset.shuffle.seed == 123
    assert custom_dataset.splits.train == 0.7
    assert custom_dataset.splits.test == 0.3
    assert custom_dataset.attributes.required_columns == ["col1"]



import copy
import inspect
from functools import lru_cache
from typing import Any, Dict, List, Set

from openai import OpenAI

from cyyrus.models.options import LargeLanguageModels, VisionLanguageModels
from cyyrus.models.task_type import TaskType
from cyyrus.models.types import DefaultModel
from cyyrus.tasks.base import BaseTask
from cyyrus.tasks.utils import Base64ImageFinder, GeneralUtils, NestedDictAccessor
from cyyrus.utils.errors import error_handler
from cyyrus.utils.logging import get_logger

logger = get_logger(__name__)

catch_all_return_none = error_handler(
    exceptions=Exception,  # This will catch all exceptions
    handler=None,
    logger=logger,
    retries=1,
    default_return=None,
)


class GenerationTask(BaseTask):
    TASK_ID = TaskType.GENERATION

    SUPPORTS_REFERENCE_FREE_EXECUTION = True
    DEFAULT_MODEL = LargeLanguageModels.GPT_4O_MINI
    DEFAULT_PROMPT = "Convert the corpus into a usable dataset"
    MAX_EPOCH = 100

    @lru_cache
    def get_valid_completion_args(self) -> Set[str]:
        completion_params = inspect.signature(self.client.chat.completions.create).parameters
        # Return a set of valid argument names
        return set(completion_params.keys())

    def __init__(
        self,
        column_name: str,
        task_properties: Dict[str, Any],
    ) -> None:
        super().__init__(
            column_name,
            task_properties,
        )
        api_key = NestedDictAccessor.get_nested_value(
            task_properties,
            "api_key",
        )
        self.client = OpenAI(
            api_key=api_key,
        )

    def execute(
        self,
        task_input: Dict[str, Any],
    ) -> Any:
        """
        Perform the generation task.
        """
        logger.debug(f"Executing generation task with {self.TASK_ID} ...")
        return self.inference(task_input=task_input)

    @catch_all_return_none
    def inference(self, task_input: Dict[str, Any]) -> Any:
        # Get Prompt from Task Property
        prompt = NestedDictAccessor.get_nested_value(
            self.task_properties,
            key="prompt",
            default=GenerationTask.DEFAULT_PROMPT,
        )

        # Process Prompt
        formatted_prompt = GeneralUtils.populate_template(
            prompt,
            self.task_properties | task_input,  # Merge task_properties and task_input
        )
        logger.debug(f"Formatted Prompt: {formatted_prompt}")

        # Pop prompt parameter
        generation_properties = copy.deepcopy(self.task_properties)
        generation_properties.pop("prompt", None)

        # Format OpenAI Call
        content = [
            {
                "type": "text",
                "text": formatted_prompt,
            }
        ]

        # Generation Model
        model = generation_properties.get(
            "model",
            LargeLanguageModels.GPT_4O_MINI,
        )

        # Incase model supports vision params
        if GenerationTask.supports_vision(model=model):
            # Find base64 encoded images in the task_input
            img_keys = Base64ImageFinder.find_base64_encoded_keys(task_input=task_input)
            logger.debug(f"Found {len(img_keys)} base64 images in task input ...")

            for img_key in img_keys:
                base64_image = NestedDictAccessor.get_nested_value(
                    task_input,
                    img_key,
                    None,
                )
                if base64_image:
                    image_arg = {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/jpeg;base64,{base64_image}",  # TODO(wizenheimer): add support for other image types
                            "detail": "low",  # TODO(wizenheimer): make this configurable
                        },
                    }
                    content.append(image_arg)

        # Create the messages list with the proper structure
        messages = [
            {
                "role": "user",
                "content": content,
            }
        ]

        # Update generation_property with the correct messages format
        generation_properties["messages"] = messages

        # Check if structured prompting is supported
        if generation_properties.get("response_format", None) is None:
            logger.debug("No response format specified, defaulting to unstructured text ...")
            generation_properties["response_format"] = DefaultModel
        else:
            logger.debug("Response format specified, attempting to process ...")

        # Filter out unnecessary properties form generation properties
        filtered_generation_property = {
            k: v for k, v in generation_properties.items() if k in self.get_valid_completion_args()
        }

        # Perform completion and handle errors
        response = self.client.beta.chat.completions.parse(**filtered_generation_property)

        # Return results
        if response.choices[0].message.refusal:
            logger.error(f"Request refused: {response.choices[0].message.refusal}")
            return None

        if generation_properties.get("response_format", None) == DefaultModel:
            return response.choices[0].message.parsed.model_dump().get("value", None)  # type: ignore

        return response.choices[0].message.parsed.model_dump()  # type: ignore

    def _generate_references(
        self,
    ) -> List[Dict[str, Any]]:
        """
        Attempt to generate the reference data for the task, using the task properties
        """
        max_epochs = NestedDictAccessor.get_nested_value(
            self.task_properties,
            "max_epochs",
            GenerationTask.MAX_EPOCH,
        )

        # Attempt to generate the reference data
        logger.debug(f"Generating references for task {self.TASK_ID} with {max_epochs} epochs ...")
        return [{} for _ in range(max_epochs)]

    @staticmethod
    def supports_vision(
        model: str,
    ) -> bool:
        """
        Check if the model supports vision parameters.
        """
        return model in VisionLanguageModels.__members__.values()



# pre_install.py

import platform
import subprocess
import sys


def run_command(command):
    try:
        result = subprocess.run(command, shell=True, text=True, capture_output=True)
        result.check_returncode()
        return result.stdout
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr.strip())


def install_package(command, package_name):
    try:
        output = run_command(command)
        print(output)
        return output
    except RuntimeError as e:
        raise RuntimeError(f"Failed to install {package_name}: {e}")


def check_and_install():
    try:
        # Check and install Poppler
        try:
            run_command("pdftoppm -h")
        except RuntimeError:
            if platform.system() == "Darwin":  # macOS
                install_package("brew install poppler", "Poppler")
            elif platform.system() == "Linux":  # Linux
                install_package(
                    "sudo apt-get update && sudo apt-get install -y poppler-utils",
                    "Poppler",
                )
            else:
                raise RuntimeError(
                    "Please install Poppler manually from https://poppler.freedesktop.org/"
                )

        # Check and install FFmpeg
        try:
            run_command("ffmpeg -version")
            print("FFmpeg is already installed.")
        except RuntimeError:
            if platform.system() == "Darwin":  # macOS
                install_package("brew install ffmpeg", "FFmpeg")
            elif platform.system() == "Linux":  # Linux
                install_package(
                    "sudo apt-get update && sudo apt-get install -y ffmpeg",
                    "FFmpeg",
                )
            else:
                raise RuntimeError(
                    "Please install FFmpeg manually from https://ffmpeg.org/download.html"
                )

    except RuntimeError as err:
        print(f"Error during installation: {err}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    check_and_install()



# import logging

from .models.spec import load_spec

# from .utils.logging import get_logger, setup_logging

# Set up logging for the entire package
# setup_logging(
#     log_level=logging.DEBUG,
#     log_file="cyyrus.log",
#     for_human=True,
# )
# logger = get_logger()

__all__ = [
    "load_spec",
]



import os
import subprocess
import sys
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.develop import develop
from setuptools.command.install import install

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


class InstallSystemDependencies(install):
    """Post-installation for installation mode."""

    def run(self):
        """Run post-install hook."""
        try:
            subprocess.check_call([sys.executable, "-m", "package.python.extras.pre_install"])
        except subprocess.CalledProcessError as e:
            print(f"Pre-install script failed: {e}", file=sys.stderr)
            sys.exit(1)
        install.run(self)


class InstallDevelopmentSystemDependencies(develop):
    """Post-installation for development mode."""

    def run(self):
        """Run post-install hook."""
        try:
            subprocess.check_call([sys.executable, "-m", "package.python.extras.pre_install"])
        except subprocess.CalledProcessError as e:
            print(f"Pre-install script failed: {e}", file=sys.stderr)
            sys.exit(1)
        install.run(self)  # type: ignore


def parse_requirements(filename):
    with open(filename, "r") as f:
        return [
            line.split("--hash=")[0].strip()
            for line in f
            if line.strip() and not line.startswith("#")
        ]


requirements_path = os.path.join("requirements", "requirements.in")
requirements = parse_requirements(requirements_path)

packages = find_packages(
    where="package/python",
    exclude=[
        # Exclude tests if present
        "*.tests",
        "*.tests.*",
        "tests.*",
        "tests",
        # Exclude __pycache__ directories
        "*.__pycache__",
        "*.__pycache__.*",
    ],
)

setup(
    name="cyyrus",
    description="Transform Unstructured Data into Usable Datasets",
    author="wizenheimer",
    version="0.14.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="cyyruslabs@gmail.com",
    url="https://github.com/wizenheimer/cyyrus",
    packages=packages,
    package_dir={
        "": "package/python",
    },
    install_requires=requirements,
    cmdclass={
        "install": InstallSystemDependencies,
        "develop": InstallDevelopmentSystemDependencies,
    },
    entry_points={
        "console_scripts": [
            "cyyrus=cyyrus.cli.main:cli",
        ],
    },
    include_package_data=True,
    exclude_package_data={
        "": [".DS_Store"],
        "package": [".DS_Store"],
        "package/python": [".DS_Store"],
        "package/python/cyyrus": [".DS_Store"],
    },
)



import os
import re
from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Dict, Generator, List, Optional, Set, Tuple, Union
from urllib.parse import urlparse

import requests
import yaml
from dotenv import load_dotenv
from pydantic import BaseModel, model_validator
from pydantic.fields import Field

from cyyrus.constants.messages import Messages
from cyyrus.models.column import Column
from cyyrus.models.dataset import Dataset, SpecVersion
from cyyrus.models.task import Task, TaskType
from cyyrus.models.types import CustomType, TypeMappingUtils
from cyyrus.utils.logging import get_logger
from cyyrus.utils.mermaid import Mermaid

logger = get_logger(__name__)


class Spec(BaseModel):
    spec: SpecVersion
    dataset: Dataset
    tasks: Dict[str, Task]
    types: Optional[Dict[str, CustomType]] = Field(
        default=None,
        exclude=True,
    )
    columns: Dict[str, Column]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._populate_concrete_model()

    def _populate_concrete_model(
        self,
    ):
        """
        Create a concrete model for the response_format specified in the task properties.
        """
        logger.debug("Attempting to populate concrete model for response_format")
        # Create a defaultdict that returns None for missing keys
        custom_types: DefaultDict[str, Optional[Dict[str, Any]]] = defaultdict(lambda: None)

        # Extract the custom types
        # Populate the defaultdict with the custom types
        if self.types:
            for type_name, custom_type in self.types.items():
                custom_types[type_name] = custom_type.model_dump()

        # Update the task properties with the concrete model, incase response_format is specified
        for task_id, task in self.tasks.items():
            # Check if the task is of type generation
            if task.task_type == TaskType.GENERATION:
                # Get the response_format_identifier
                response_format_identifier = task.task_properties.get(
                    "response_format",
                    None,
                )

                # Skip if response_format is not specified
                if not response_format_identifier:
                    continue

                # Get the concrete model for the response_format
                concrete_type_def = custom_types.get(
                    response_format_identifier,
                )
                concrete_model = TypeMappingUtils.get_concrete_model(
                    concrete_type_def,
                )
                task.task_properties["response_format"] = concrete_model
                logger.debug(
                    f"Populated concrete model for response_format {response_format_identifier}"
                )

                # Update the column properties with the concrete model
                self.tasks[task_id] = task
                logger.debug(f"Updated task {task_id} with concrete model")

        logger.debug("Populated concrete models for response_format")

    def extract_dag_representation(
        self,
    ) -> Dict[
        str,
        List[str],
    ]:
        """
        Extract the DAG representation of the spec.
        """
        return {
            column_name: list(column.task_input) for column_name, column in self.columns.items()
        }

    @model_validator(mode="after")
    def validate_dag(
        cls,
        values,
    ):
        """
        Validate the DAG representation of the spec.
        """
        logger.debug("Validating the DAG representation of the spec")
        dag = values.extract_dag_representation()

        def has_cycle(node: str, visited: Set[str], rec_stack: Set[str]) -> bool:
            visited.add(node)
            rec_stack.add(node)

            for neighbor in dag.get(node, []):
                if neighbor not in visited:
                    if has_cycle(neighbor, visited, rec_stack):
                        return True
                elif neighbor in rec_stack:
                    return True

            rec_stack.remove(node)
            return False

        visited: Set[str] = set()
        rec_stack: Set[str] = set()

        for node in dag:
            if node not in visited:
                logger.debug(f"Checking for cycle starting at `{node}` node")
                if has_cycle(node, visited, rec_stack):
                    logger.error(f"{Messages.TASK_INPUTS_CYCLIC_DEPENDENCY}: {node}")
                    raise ValueError(Messages.TASK_INPUTS_CYCLIC_DEPENDENCY)

        logger.debug("DAG representation is valid")
        return values

    @model_validator(mode="after")
    def validate_columns_for_orphans(
        cls,
        values,
    ):
        """
        Validate that all columns have a task associated with them.
        """
        logger.debug("Validating columns for orphans")
        # Note: duplicates are handled by Pydantic + YAML Parser
        all_column_names = set(values.columns.keys())

        # Check if all required columns exist in either columns or references
        required_columns = set(values.dataset.attributes.required_columns)
        missing_columns = required_columns - all_column_names
        if missing_columns:
            logger.error(f"{Messages.REQUIRED_COLUMN_MISSING} columns: {missing_columns}")
            raise ValueError(Messages.REQUIRED_COLUMN_MISSING)

        # Check if all flatten columns exist in either columns or references
        flatten_columns = set(values.dataset.attributes.flatten_columns)
        missing_columns = flatten_columns - all_column_names
        if missing_columns:
            logger.error(f"{Messages.REQUIRED_COLUMN_MISSING} columns: {missing_columns}")
            raise ValueError(Messages.REQUIRED_COLUMN_MISSING)

        # Check if all unique columns exist in either columns or references
        unique_columns = set(values.dataset.attributes.unique_columns)
        missing_columns = unique_columns - all_column_names
        if missing_columns:
            logger.error(f"{Messages.REQUIRED_COLUMN_MISSING} columns: {missing_columns}")
            raise ValueError(Messages.REQUIRED_COLUMN_MISSING)

        # Check if all exclude columns exist in either columns or references
        excluded_columns = set(values.dataset.attributes.exclude_columns)
        missing_columns = excluded_columns - all_column_names
        if missing_columns:
            logger.error(f"{Messages.REQUIRED_COLUMN_MISSING} columns: {missing_columns}")
            raise ValueError(Messages.REQUIRED_COLUMN_MISSING)

        logger.debug("Columns are valid")
        return values

    @model_validator(mode="after")
    def validate_task_ids(
        cls,
        values,
    ):
        """
        Validate that all columns have a valid task_id associated with them.
        """
        logger.debug("Validating task IDs")
        tasks = values.tasks
        for column_name, column in values.columns.items():
            if column.task_id not in tasks:
                logger.error(
                    f"{Messages.COLUMN_TASK_ID_NOT_FOUND}: {column.task_id} for {column_name}"
                )
                raise ValueError(Messages.COLUMN_TASK_ID_NOT_FOUND)
        logger.debug("Task IDs are validated")
        return values

    def extract_task_info(
        self,
        column_name: str,
    ) -> Tuple[
        List[str],
        str,
        TaskType,
        Dict[str, Union[int, str, float]],
    ]:
        """
        Extract the task information for a given column.
        """
        column = self.columns.get(column_name)

        if not column:
            # If the column is still not found, raise an error
            logger.error(f"{Messages.COLUMNN_ID_NOT_FOUND}: {column_name}")
            raise ValueError(Messages.COLUMNN_ID_NOT_FOUND)

        task = self.tasks.get(column.task_id)
        if not task:
            logger.error(f"{Messages.COLUMN_TASK_ID_NOT_FOUND}: {column_name}")
            raise ValueError(Messages.COLUMN_TASK_ID_NOT_FOUND)

        return (
            # Column Information
            column.task_input,
            column_name,
            # Task Information
            task.task_type,
            task.task_properties,
        )

    def to_dict(
        self,
        *args,
        **kwargs,
    ):
        """
        Convert the Spec object to a dictionary.
        """
        return super().model_dump(*args, **kwargs)

    def preview(
        self,
    ) -> Mermaid:
        """
        Generate a Mermaid graph representation of the DAG.
        """
        logger.debug("Generating a Mermaid graph representation of the DAG")
        dag = self.extract_dag_representation()
        mermaid_code = ["graph LR"]
        for task, dependencies in dag.items():
            for dep in dependencies:
                mermaid_code.append(f"    {dep} --> {task}")
        mermaid_code = "\n".join(mermaid_code)
        return Mermaid(mermaid_code)

    def levels(
        self,
    ) -> Generator[
        List[
            Tuple[
                List[str],
                str,
                TaskType,
                Dict[str, Any],
            ]
        ],
        None,
        None,
    ]:
        """
        Generate the levels of the DAG.
        """
        # Extract the DAG representation
        dependencies: Dict[str, List[str]] = self.extract_dag_representation()

        for level in level_order_traversal(dependencies):
            yield [self.extract_task_info(node) for node in level]


def env_var_constructor(
    loader,
    node,
):
    """
    Custom YAML constructor to replace environment variables.
    """
    value = loader.construct_scalar(node)

    # Handle ${VAR} syntax
    match = re.match(r"\${(.+)}", value)
    if match:
        env_var = match.group(1)
    else:
        # Handle $VAR syntax, allowing for whitespace after $
        env_var = value.strip().lstrip("$").strip()

    # Look up the environment variable, fall back to the original value if not found
    return os.environ.get(env_var, value)


def load_spec(
    path_or_url: str,
    env_file: Optional[str] = None,
):
    """
    Load a spec from either a local file path or a URL, with support for environment variables.
    """
    # Load environment variables from .env file if specified
    if env_file:
        logger.debug(f"Loading environment variables from {env_file}")
        load_dotenv(env_file)

    # Check if the source is a URL
    parsed_url = urlparse(path_or_url)
    if parsed_url.scheme and parsed_url.netloc:
        # It's a URL, so use requests to fetch the content
        try:
            logger.info(f"Fetching spec from {path_or_url}")
            response = requests.get(path_or_url)
            response.raise_for_status()  # Raise an exception for HTTP errors
            yaml_content = response.text
        except requests.RequestException as e:
            logger.error(f"{Messages.SCHEMA_FILE_COULD_NOT_BE_LOCATED} {path_or_url}: {e}")
            raise ValueError(Messages.SCHEMA_FILE_COULD_NOT_BE_LOCATED)
    elif Path(path_or_url).is_file():
        # It's a local file, so read its content
        try:
            logger.info(f"Reading spec from {path_or_url}")
            with open(path_or_url, "r") as file:
                yaml_content = file.read()
        except IOError as e:
            logger.error(f"{Messages.SCHEMA_FILE_COULD_NOT_BE_LOCATED} {path_or_url}: {e}")
            raise ValueError(Messages.SCHEMA_FILE_COULD_NOT_BE_LOCATED)
    else:
        # The path is neither a URL nor a file
        logger.error(
            f"{Messages.SCHEMA_FILE_COULD_NOT_BE_LOCATED} unrecognized path: {path_or_url}"
        )
        raise ValueError(Messages.SCHEMA_FILE_COULD_NOT_BE_LOCATED)

    # Parse the YAML content
    try:
        # Use a custom loader that expands environment variables
        class EnvVarLoader(yaml.SafeLoader):
            pass

        # Add implicit resolver for environment variables
        EnvVarLoader.add_implicit_resolver(
            "env_var",
            re.compile(r"(\$\s*[A-Za-z_][A-Za-z0-9_]*|\$\{[A-Za-z_][A-Za-z0-9_]*\})"),
            first=["$"],
        )

        # Add constructor for environment variables
        EnvVarLoader.add_constructor("env_var", env_var_constructor)

        # Parse the YAML content
        logger.info("Parsing the spec")
        parsed_spec = Spec(**yaml.load(yaml_content, Loader=EnvVarLoader))

        logger.info("Parsed and validated the spec")
        return parsed_spec
    except yaml.YAMLError as e:
        # Raise an error if the YAML parsing fails
        logger.error(f"{Messages.SCHEMA_COULD_NOT_BE_PARSED}: {e}")
        raise ValueError(Messages.SCHEMA_COULD_NOT_BE_PARSED)


def level_order_traversal(
    dependencies: Dict[
        str,
        List[str],
    ],
):
    """
    Perform a level order traversal of the DAG.
    """
    logger.debug("Performing level order traversal of the DAG")
    # Create a reverse dependency graph using defaultdict
    reverse_deps = defaultdict(list)
    all_nodes = set()

    for node, deps in dependencies.items():
        all_nodes.add(node)
        all_nodes.update(deps)
        for dep in deps:
            reverse_deps[dep].append(node)

    # Ensure all nodes are in the dependencies dict
    for node in all_nodes:
        if node not in dependencies:
            dependencies[node] = []

    # Find nodes with no dependencies (root nodes)
    root_nodes = [node for node in all_nodes if not dependencies[node]]

    visited = set()
    current_level = root_nodes

    while current_level:
        current_level = list(dict.fromkeys(current_level))  # Remove duplicates
        visited.update(current_level)
        next_level = []
        for node in current_level:
            for child in reverse_deps[node]:
                if child not in visited and all(dep in visited for dep in dependencies[child]):
                    # Add the child to the next level
                    next_level.append(child)
        # Yield the current level
        yield current_level
        current_level = next_level



from enum import Enum

import pytest
from cyyrus.errors.column import (  # type: ignore
    ColumnIDNotFoundError,
    ColumnTaskIDNotFoundError,
    ColumnTypeNotFoundError,
)
from cyyrus.models.column import Column  # type: ignore
from cyyrus.models.dataset import (  # type: ignore
    Dataset,
    DatasetAttributes,
    SpecVersion,
)
from cyyrus.models.spec import Spec  # type: ignore
from cyyrus.models.task import Task  # type: ignore
from cyyrus.models.task_type import TaskType  # type: ignore
from cyyrus.models.types import (  # type: ignore
    CustomType,
    DataType,
)
from pydantic import ValidationError


def test_missing_field():
    invalid_schema = {
        # Missing required fields
        "spec": "v0",
        "tasks": {},
        "types": {},
    }

    with pytest.raises(ValidationError):
        Spec.model_validate(invalid_schema)


def test_spec_version():
    assert SpecVersion.V0 == "v0"
    assert isinstance(SpecVersion.V0, Enum)


def test_spec():
    # Create mock data for testing
    with pytest.warns():
        mock_dataset = Dataset()
    mock_tasks = {"task1": Task(task_type=TaskType.DEFAULT, task_properties={})}
    mock_types = {"type1": CustomType(type=DataType.STRING)}
    mock_columns = {"col1": Column(column_type="string", task_id="task1")}

    # Test valid Spec
    valid_spec = Spec(
        spec=SpecVersion.V0,
        dataset=mock_dataset,
        tasks=mock_tasks,
        types=mock_types,
        columns=mock_columns,
    )
    assert valid_spec.spec == SpecVersion.V0
    assert isinstance(valid_spec.dataset, Dataset)
    assert "task1" in valid_spec.tasks
    assert "type1" in valid_spec.types
    assert "col1" in valid_spec.columns

    # Test invalid Spec (missing required column)
    with pytest.raises(ColumnIDNotFoundError):
        with pytest.warns():
            dataset = Dataset(attributes=DatasetAttributes(required_columns=["missing_column"]))
        Spec(
            spec=SpecVersion.V0,
            dataset=dataset,
            tasks=mock_tasks,
            types=mock_types,
            columns=mock_columns,
        )

    # Test invalid Spec (column type not found)
    with pytest.raises(ColumnTypeNotFoundError):
        Spec(
            spec=SpecVersion.V0,
            dataset=mock_dataset,
            tasks=mock_tasks,
            types=mock_types,
            columns={"invalid_col": Column(column_type="invalid_type", task_id="task1")},
        )

    # Test invalid Spec (column task ID not found)
    with pytest.raises(ColumnTaskIDNotFoundError):
        Spec(
            spec=SpecVersion.V0,
            dataset=mock_dataset,
            tasks=mock_tasks,
            types=mock_types,
            columns={"invalid_col": Column(column_type="string", task_id="invalid_task")},
        )



# package/python/cli/main.py
import logging
import os
from pathlib import Path
from typing import Optional

import click
from pandas import DataFrame

from cyyrus.cli.utils import create_export_filepath, get_ascii_art
from cyyrus.cli.visualizer import Visualizer
from cyyrus.composer.core import Composer
from cyyrus.composer.dataframe import ExportFormat
from cyyrus.composer.utils import FunnyBones
from cyyrus.models.spec import Spec, load_spec
from cyyrus.utils.logging import get_logger, setup_logging


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(
        [
            "DEBUG",
            "INFO",
            "WARNING",
            "ERROR",
            "CRITICAL",
        ]
    ),
    help="Set the logging level",
)
@click.option(
    "--human-readable",
    is_flag=True,
    default=True,
    help="Use human-readable log format (default: True)",
)
@click.option(
    "--log-file",
    default="cyyrus.log",
    help="Name of the log file",
)
@click.option(
    "--log-dir",
    type=click.Path(path_type=Path),
    default=None,
    help="Directory to store log files",
)
@click.option(
    "--schema-path",
    type=click.Path(exists=True),
    required=True,
    help="Path to the schema file",
)
@click.option(
    "--env-path",
    type=click.Path(exists=True),
    help="Path to the optional environment file",
)
@click.option(
    "--export-format",
    type=click.Choice([format.value for format in ExportFormat]),
    default=None,
    help="Format to export the dataset",
)
@click.option(
    "--export-path",
    type=click.Path(path_type=Path),
    default=None,
    help="Directory to export the dataset",
)
@click.option(
    "--huggingface-token",
    type=str,
    default=None,
    help="Hugging Face token",
)
@click.option(
    "--repo-id",
    type=str,
    default=None,
    help="Hugging Face repo ID",
)
def run(
    log_level,
    human_readable,
    log_file,
    log_dir,
    schema_path,
    env_path,
    export_format,
    export_path,
    huggingface_token,
    repo_id,
):
    # ============================
    #       Setup logging
    # ============================
    if log_dir:
        log_file = Path(log_dir) / log_file

    setup_logging(
        log_level=getattr(logging, log_level.upper()),
        log_file=str(log_file),
        log_dir=log_dir,
        for_human=human_readable,
    )
    logger = get_logger(__name__)

    # ============================
    #      Prepare the CLI
    # ============================
    print(get_ascii_art())
    logger.info("CLI started. Buckle up, it's going to be a wild wild ride!")

    configure_hf()

    # ============================
    #      Load the spec
    # ============================
    spec = load_spec(schema_path, env_path)
    logger.info("Spec loaded. Ready to roll!")

    # ============================
    #      Initialize the composer
    # ============================
    composer = Composer(spec=spec)

    # ============================
    #      Perform the dry run
    # ============================
    if perform_dry_run():
        composer.compose(dry_run=True)
        logger.info("Dry run complete. Nothing exploded, good start!")
    else:
        logger.info("Skipping dry run. Brave choice")

    # ========================================================
    #      Perform the full run and display the sample dataframe
    # ========================================================
    if perform_full_run():
        composer.compose()
        logger.info("Full run executed. Hope you liked it!")
    else:
        logger.info("Skipping full run. Until next time!")
        return

    display_intermediate_results(
        composer.dataframe,
        spec,
        logger,
    )

    # ============================
    #      Export the dataset
    # ============================
    if export_dataset(
        composer=composer,
        export_path=export_path,
        export_format=export_format,
        logger=logger,
    ):
        if publish_dataset(
            composer=composer,
            huggingface_token=huggingface_token,
            repo_id=repo_id,
            logger=logger,
            export_path=export_path,
            export_format=export_format,
        ):
            logger.info("Dataset published successfully!")
        else:
            logger.info("Dataset publishing skipped or failed.")
    else:
        logger.info("Dataset export skipped.")


def configure_hf():
    os.environ["HF_HUB_DISABLE_EXPERIMENTAL_WARNING"] = "1"


def perform_dry_run():
    return click.confirm("Do you want to perform a dry run?", default=True)


def perform_full_run():
    return click.confirm("Do you want to perform a full run?", default=True)


def display_intermediate_results(
    df: DataFrame,
    spec: Spec,
    logger,
):
    logger.info("Here's a sneak peek of your data. Doesn't it look fabulous?")
    Visualizer.display_dataframe_properties(
        df,
    )
    Visualizer.display_dataset_properties(
        spec.dataset,
    )


def export_dataset(
    composer: Composer,
    export_path: Path,
    export_format: ExportFormat,
    logger: logging.Logger,
):
    if not click.confirm("Ready to export the dataset?"):
        return False

    while True:
        export_path, export_format, dataset_name = prompt_export_details(
            export_path,
            export_format,
        )
        full_export_path = create_export_filepath(
            Path(export_path),
            dataset_name,
            export_format,
        )

        if full_export_path.exists() and not click.confirm(
            f"File {full_export_path} already exists. Overwrite?"
        ):
            continue

        try:
            composer.export(export_format=export_format, filepath=full_export_path)
            logger.info(
                f"Exported dataset to {export_path} in {export_format} format. It's now officially your data!"
            )
            return True
        except Exception as e:
            logger.error(f"Export failed due to: {str(e)}")
            if not click.confirm(
                "Do you want to try exporting again with a different format or path?"
            ):
                logger.info("Skipping export. Moving on!")
                return False


def prompt_export_details(export_path, export_format):
    if export_path is None:
        export_path = click.prompt(
            "Enter the export directory",
            type=click.Path(path_type=Path),
            default=Path.cwd(),
        )

    if export_format is None:
        export_format = click.prompt(
            "Enter the export format",
            type=click.Choice([format.value for format in ExportFormat]),
            default=ExportFormat.HUGGINGFACE.value,
        )

    suggested_name = FunnyBones.suggest()
    dataset_name = click.prompt(
        f"Enter a name for your dataset (How about: {suggested_name} ?)",
        default=suggested_name,
    )

    return export_path, export_format, dataset_name


def publish_dataset(
    composer: Composer,
    logger: logging.Logger,
    export_path: Path,
    export_format: ExportFormat,
    huggingface_token: Optional[str] = None,
    repo_id: Optional[str] = None,
):
    if not click.confirm("Do you want to publish the dataset?"):
        return False

    hf_token = get_huggingface_token(huggingface_token)
    repository_id: str = (
        click.prompt("Enter the repository identifier", type=str) if repo_id is None else repo_id
    )
    is_private = click.confirm("Keep the dataset private?", default=False)

    try:
        composer.publish(
            repository_id=repository_id,
            huggingface_token=hf_token,
            private=is_private,
        )
        logger.info(f"Published dataset to {repo_id}. Happy sharing!")
        return True
    except Exception as e:
        logger.error(f"Publishing failed due to: {str(e)}")
        if click.confirm("Do you want to retry publishing?"):
            try:
                composer.publish(
                    repository_id=repository_id,
                    huggingface_token=hf_token,
                    private=is_private,
                )
                logger.info(f"Published dataset to {repo_id}. Happy sharing!")
                return True
            except Exception as e:
                logger.error(f"Publishing failed again due to: {str(e)}")
                return (
                    export_dataset(
                        composer=composer,
                        logger=logger,
                        export_format=export_format,
                        export_path=export_path,
                    )
                    if click.confirm("Do you want to export the dataset locally instead?")
                    else False
                )
        return False


def get_huggingface_token(huggingface_token):
    if huggingface_token:
        return huggingface_token

    hf_token = os.environ.get("HF_TOKEN")
    if hf_token and click.confirm(
        f"HF_TOKEN found in environment. Use '{hf_token[:5]}...{hf_token[-5:]}'?"
    ):
        return hf_token

    return click.prompt("Enter the Hugging Face token", type=str, hide_input=True)


def handle_publishing_error(composer, error, repo_id, hf_token, is_private, logger):
    logger.error(f"Publishing failed due to: {str(error)}")
    if click.confirm("Do you want to retry publishing?"):
        try:
            composer.publish(repo_id=repo_id, huggingface_token=hf_token, private=is_private)
            logger.info(f"Published dataset to {repo_id}. Happy sharing!")
            return True
        except Exception as e:
            logger.error(f"Publishing failed again due to: {str(e)}")
            return (
                handle_local_export(composer, logger)
                if click.confirm("Do you want to export the dataset locally instead?")
                else False
            )
    return False


def handle_local_export(composer, logger):
    export_path = click.prompt(
        "Enter the export directory",
        type=click.Path(path_type=Path),
        default=Path.cwd(),
    )
    export_format = click.prompt(
        "Enter the export format",
        type=click.Choice([format.value for format in ExportFormat]),
        default=ExportFormat.HUGGINGFACE.value,
    )
    full_export_path = create_export_filepath(
        Path(export_path),
        click.prompt("Enter a name for your dataset", type=str),
        export_format,
    )
    composer.export(export_format=export_format, filepath=full_export_path)
    logger.info(f"Exported dataset to {full_export_path} in {export_format} format.")
    return True


if __name__ == "__main__":
    cli()
